# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict

from sqlalchemy import Boolean, Column, String, Table

from .._repository._base_taipy_model import _BaseModel
from .._repository.db._sql_base_model import mapper_registry


@mapper_registry.mapped
@dataclass
class _VersionModel(_BaseModel):
    __table__ = Table(
        "version",
        mapper_registry.metadata,
        Column("id", String, primary_key=True),
        Column("config", String),  # config is store as a json string
        Column("creation_date", String),
        Column("is_production", Boolean),
        Column("is_development", Boolean),
        Column("is_latest", Boolean),
    )
    id: str
    config: Dict[str, Any]
    creation_date: str

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        return _VersionModel(
            id=data["id"],
            config=data["config"],
            creation_date=data["creation_date"],
        )



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict

from sqlalchemy import JSON, Column, Enum, String, Table

from taipy.config.common.frequency import Frequency

from .._repository._base_taipy_model import _BaseModel
from .._repository.db._sql_base_model import mapper_registry
from .cycle_id import CycleId


@mapper_registry.mapped
@dataclass
class _CycleModel(_BaseModel):
    __table__ = Table(
        "cycle",
        mapper_registry.metadata,
        Column("id", String, primary_key=True),
        Column("name", String),
        Column("frequency", Enum(Frequency)),
        Column("properties", JSON),
        Column("creation_date", String),
        Column("start_date", String),
        Column("end_date", String),
    )
    id: CycleId
    name: str
    frequency: Frequency
    properties: Dict[str, Any]
    creation_date: str
    start_date: str
    end_date: str

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        return _CycleModel(
            id=data["id"],
            name=data["name"],
            frequency=Frequency._from_repr(data["frequency"]),
            properties=data["properties"],
            creation_date=data["creation_date"],
            start_date=data["start_date"],
            end_date=data["end_date"],
        )



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""# Taipy Core

The Taipy Core package is a Python library designed to build powerful, customized, data-driven back-end
applications. It provides the tools to help Python developers transform their algorithms into a
complete back-end application.
More details on the [Taipy Core](../../core/index.md) functionalities are available in the user manual.

To build a Taipy Core application, the first step consists of setting up the Taipy configuration to design
your application's characteristics and behaviors.
Import `Config^` from the `taipy.config^` module, then use the various methods of the `Config^` singleton class to
configure your core application. In particular, configure the
[data nodes](../../core/config/data-node-config.md), [tasks](../../core/config/task-config.md),
and [scenarios](../../core/config/scenario-config.md).
Please refer to the [Core configuration user manual](../../core/config/index.md) for more information and
detailed examples.

Once your application is configured, import module `import taipy as tp` so you can use any function described
in the following section on [Functionc](#functions). In particular, the most used functions are `tp.create_scenario()`,
`tp.get_scenarios()`, `tp.get_data_nodes()`, `tp.submit()`, used to get, create, and submit entities.

!!! Note

    Taipy Core provides a runnable service, `Core^` that runs as a service in a dedicated thread. The purpose is to
    have a dedicated thread responsible for dispatching the submitted jobs to an available executor for their execution.

    In particular, this `Core^` service is automatically run when Core is used with Taipy REST or Taipy GUI. See the
    [running services](../../run-deploy/run/running_services.md) page of the user manual for more details.
"""

from ._init import *
from ._init_version import _read_version
from .common.mongo_default_document import MongoDefaultDocument
from .data.data_node_id import Edit
from .exceptions import exceptions

__version__ = _read_version()



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import uuid
from typing import Callable, Iterable, List, Optional, Union

from .._manager._manager import _Manager
from .._repository._abstract_repository import _AbstractRepository
from .._version._version_manager_factory import _VersionManagerFactory
from .._version._version_mixin import _VersionMixin
from ..exceptions.exceptions import JobNotDeletedException
from ..notification import EventEntityType, EventOperation, _publish_event
from ..task.task import Task
from .job import Job
from .job_id import JobId


class _JobManager(_Manager[Job], _VersionMixin):

    _ENTITY_NAME = Job.__name__
    _ID_PREFIX = "JOB_"
    _repository: _AbstractRepository

    @classmethod
    def _get_all(cls, version_number: Optional[str] = None) -> List[Job]:
        """
        Returns all entities.
        """
        filters = cls._build_filters_with_version(version_number)
        return cls._repository._load_all(filters)

    _EVENT_ENTITY_TYPE = EventEntityType.JOB

    @classmethod
    def _create(
        cls, task: Task, callbacks: Iterable[Callable], submit_id: str, submit_entity_id: str, force=False
    ) -> Job:
        version = _VersionManagerFactory._build_manager()._get_latest_version()
        job = Job(
            id=JobId(f"{Job._ID_PREFIX}_{task.config_id}_{uuid.uuid4()}"),
            task=task,
            submit_id=submit_id,
            submit_entity_id=submit_entity_id,
            force=force,
            version=version,
        )
        cls._set(job)
        _publish_event(cls._EVENT_ENTITY_TYPE, job.id, EventOperation.CREATION, None)
        job._on_status_change(*callbacks)
        return job

    @classmethod
    def _delete(cls, job: Job, force=False):
        if job.is_finished() or force:
            super()._delete(job.id)
            from .._orchestrator._dispatcher._job_dispatcher import _JobDispatcher

            _JobDispatcher._pop_dispatched_process(job.id)
        else:
            err = JobNotDeletedException(job.id)
            cls._logger.warning(err)
            raise err

    @classmethod
    def _cancel(cls, job: Union[str, Job]):
        job = cls._get(job) if isinstance(job, str) else job

        from .._orchestrator._orchestrator_factory import _OrchestratorFactory

        _OrchestratorFactory._build_orchestrator().cancel_job(job)

    @classmethod
    def _get_latest(cls, task: Task) -> Optional[Job]:
        jobs_of_task = list(filter(lambda job: task in job, cls._get_all()))
        if len(jobs_of_task) == 0:
            return None
        if len(jobs_of_task) == 1:
            return jobs_of_task[0]
        else:
            return max(jobs_of_task)

    @classmethod
    def _is_deletable(cls, job: Union[Job, JobId]) -> bool:
        if isinstance(job, str):
            job = cls._get(job)
        if job.is_finished():
            return True
        return False



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime
from typing import List

from .._repository._abstract_converter import _AbstractConverter
from ..common._utils import _fcts_to_dict, _load_fct
from ..exceptions import InvalidSubscriber
from ..job._job_model import _JobModel
from ..job.job import Job
from ..task._task_manager_factory import _TaskManagerFactory


class _JobConverter(_AbstractConverter):
    @classmethod
    def _entity_to_model(cls, job: Job) -> _JobModel:
        return _JobModel(
            job.id,
            job._task.id,
            job._status,
            job._force,
            job.submit_id,
            job.submit_entity_id,
            job._creation_date.isoformat(),
            cls.__serialize_subscribers(job._subscribers),
            job._stacktrace,
            version=job._version,
        )

    @classmethod
    def _model_to_entity(cls, model: _JobModel) -> Job:
        task_manager = _TaskManagerFactory._build_manager()
        task_repository = task_manager._repository

        job = Job(
            id=model.id,
            task=task_repository._load(model.task_id),
            submit_id=model.submit_id,
            submit_entity_id=model.submit_entity_id,
            version=model.version,
        )

        job._status = model.status  # type: ignore
        job._force = model.force  # type: ignore
        job._creation_date = datetime.fromisoformat(model.creation_date)  # type: ignore
        for it in model.subscribers:
            try:
                fct_module, fct_name = it.get("fct_module"), it.get("fct_name")
                job._subscribers.append(_load_fct(fct_module, fct_name))  # type: ignore
            except AttributeError:
                raise InvalidSubscriber(f"The subscriber function {it.get('fct_name')} cannot be loaded.")
        job._stacktrace = model.stacktrace

        return job

    @staticmethod
    def __serialize_subscribers(subscribers: List) -> List:
        return _fcts_to_dict(subscribers)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ..exceptions.exceptions import InvalidExposedType


class _AbstractTabularDataNode(object):
    """Abstract base class for tabular data node implementations (CSVDataNode, ParquetDataNode, ExcelDataNode,
    SQLTableDataNode and SQLDataNode) that are tabular representable."""

    @staticmethod
    def _check_exposed_type(exposed_type, valid_string_exposed_types):
        if isinstance(exposed_type, str) and exposed_type not in valid_string_exposed_types:
            raise InvalidExposedType(
                f"Invalid string exposed type {exposed_type}. Supported values are "
                f"{', '.join(valid_string_exposed_types)}"
            )



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import datetime
from functools import partial
from typing import Any, Callable, List, Optional, Union

from taipy.config import Config

from .._entity._entity_ids import _EntityIds
from .._manager._manager import _Manager
from .._repository._abstract_repository import _AbstractRepository
from .._version._version_mixin import _VersionMixin
from ..common.warn_if_inputs_not_ready import _warn_if_inputs_not_ready
from ..config.scenario_config import ScenarioConfig
from ..cycle._cycle_manager_factory import _CycleManagerFactory
from ..cycle.cycle import Cycle
from ..data._data_manager_factory import _DataManagerFactory
from ..exceptions.exceptions import (
    DeletingPrimaryScenario,
    DifferentScenarioConfigs,
    DoesNotBelongToACycle,
    InsufficientScenarioToCompare,
    InvalidSequence,
    InvalidSscenario,
    NonExistingComparator,
    NonExistingScenario,
    NonExistingScenarioConfig,
    SequenceTaskConfigDoesNotExistInSameScenarioConfig,
    UnauthorizedTagError,
)
from ..job._job_manager_factory import _JobManagerFactory
from ..job.job import Job
from ..notification import EventEntityType, EventOperation, _publish_event
from ..task._task_manager_factory import _TaskManagerFactory
from .scenario import Scenario
from .scenario_id import ScenarioId


class _ScenarioManager(_Manager[Scenario], _VersionMixin):
    _AUTHORIZED_TAGS_KEY = "authorized_tags"
    _ENTITY_NAME = Scenario.__name__
    _EVENT_ENTITY_TYPE = EventEntityType.SCENARIO

    _repository: _AbstractRepository

    @classmethod
    def _get_all(cls, version_number: Optional[str] = None) -> List[Scenario]:
        """
        Returns all entities.
        """
        filters = cls._build_filters_with_version(version_number)
        return cls._repository._load_all(filters)

    @classmethod
    def _subscribe(
        cls,
        callback: Callable[[Scenario, Job], None],
        params: Optional[List[Any]] = None,
        scenario: Optional[Scenario] = None,
    ):
        if scenario is None:
            scenarios = cls._get_all()
            for scn in scenarios:
                cls.__add_subscriber(callback, params, scn)
            return

        cls.__add_subscriber(callback, params, scenario)

    @classmethod
    def _unsubscribe(
        cls,
        callback: Callable[[Scenario, Job], None],
        params: Optional[List[Any]] = None,
        scenario: Optional[Scenario] = None,
    ):
        if scenario is None:
            scenarios = cls._get_all()
            for scn in scenarios:
                cls.__remove_subscriber(callback, params, scn)
            return

        cls.__remove_subscriber(callback, params, scenario)

    @classmethod
    def __add_subscriber(cls, callback, params, scenario: Scenario):
        scenario._add_subscriber(callback, params)
        _publish_event(cls._EVENT_ENTITY_TYPE, scenario.id, EventOperation.UPDATE, "subscribers")

    @classmethod
    def __remove_subscriber(cls, callback, params, scenario: Scenario):
        scenario._remove_subscriber(callback, params)
        _publish_event(cls._EVENT_ENTITY_TYPE, scenario.id, EventOperation.UPDATE, "subscribers")

    @classmethod
    def _create(
        cls,
        config: ScenarioConfig,
        creation_date: Optional[datetime.datetime] = None,
        name: Optional[str] = None,
    ) -> Scenario:
        _task_manager = _TaskManagerFactory._build_manager()
        _data_manager = _DataManagerFactory._build_manager()

        scenario_id = Scenario._new_id(str(config.id))
        cycle = (
            _CycleManagerFactory._build_manager()._get_or_create(config.frequency, creation_date)
            if config.frequency
            else None
        )
        cycle_id = cycle.id if cycle else None
        tasks = (
            _task_manager._bulk_get_or_create(config.task_configs, cycle_id, scenario_id) if config.task_configs else []
        )
        additional_data_nodes = (
            _data_manager._bulk_get_or_create(config.additional_data_node_configs, cycle_id, scenario_id)
            if config.additional_data_node_configs
            else {}
        )

        sequences = {}
        tasks_and_config_id_maps = {task.config_id: task for task in tasks}
        for sequence_name, sequence_task_configs in config.sequences.items():
            sequence_tasks = []
            non_existing_sequence_task_config_in_scenario_config = set()
            for sequence_task_config in sequence_task_configs:
                if task := tasks_and_config_id_maps.get(sequence_task_config.id):
                    sequence_tasks.append(task)
                else:
                    non_existing_sequence_task_config_in_scenario_config.add(sequence_task_config.id)
            if len(non_existing_sequence_task_config_in_scenario_config) > 0:
                raise SequenceTaskConfigDoesNotExistInSameScenarioConfig(
                    list(non_existing_sequence_task_config_in_scenario_config), sequence_name, str(config.id)
                )
            sequences[sequence_name] = {Scenario._SEQUENCE_TASKS_KEY: sequence_tasks}

        is_primary_scenario = len(cls._get_all_by_cycle(cycle)) == 0 if cycle else False
        props = config._properties.copy()
        if name:
            props["name"] = name
        version = cls._get_latest_version()

        scenario = Scenario(
            config_id=str(config.id),
            tasks=set(tasks),
            properties=props,
            additional_data_nodes=set(additional_data_nodes.values()),
            scenario_id=scenario_id,
            creation_date=creation_date,
            is_primary=is_primary_scenario,
            cycle=cycle,
            version=version,
            sequences=sequences,
        )

        for task in tasks:
            if scenario_id not in task._parent_ids:
                task._parent_ids.update([scenario_id])
                _task_manager._set(task)

        for dn in additional_data_nodes.values():
            if scenario_id not in dn._parent_ids:
                dn._parent_ids.update([scenario_id])
                _data_manager._set(dn)

        cls._set(scenario)

        if not scenario._is_consistent():
            raise InvalidSscenario(scenario.id)

        actual_sequences = scenario._get_sequences()
        for sequence_name in sequences.keys():
            if not actual_sequences[sequence_name]._is_consistent():
                raise InvalidSequence(actual_sequences[sequence_name].id)
            _publish_event(EventEntityType.SEQUENCE, actual_sequences[sequence_name].id, EventOperation.CREATION, None)

        _publish_event(cls._EVENT_ENTITY_TYPE, scenario.id, EventOperation.CREATION, None)
        return scenario

    @classmethod
    def _is_submittable(cls, scenario: Union[Scenario, ScenarioId]) -> bool:
        if isinstance(scenario, str):
            scenario = cls._get(scenario)
        return isinstance(scenario, Scenario) and scenario.is_ready_to_run()

    @classmethod
    def _submit(
        cls,
        scenario: Union[Scenario, ScenarioId],
        callbacks: Optional[List[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
        check_inputs_are_ready: bool = True,
    ) -> List[Job]:
        scenario_id = scenario.id if isinstance(scenario, Scenario) else scenario
        scenario = cls._get(scenario_id)
        if scenario is None:
            raise NonExistingScenario(scenario_id)
        callbacks = callbacks or []
        scenario_subscription_callback = cls.__get_status_notifier_callbacks(scenario) + callbacks
        if check_inputs_are_ready:
            _warn_if_inputs_not_ready(scenario.get_inputs())

        jobs = (
            _TaskManagerFactory._build_manager()
            ._orchestrator()
            .submit(scenario, callbacks=scenario_subscription_callback, force=force, wait=wait, timeout=timeout)
        )
        _publish_event(cls._EVENT_ENTITY_TYPE, scenario.id, EventOperation.SUBMISSION, None)
        return jobs

    @classmethod
    def __get_status_notifier_callbacks(cls, scenario: Scenario) -> List:
        return [partial(c.callback, *c.params, scenario) for c in scenario.subscribers]

    @classmethod
    def _get_primary(cls, cycle: Cycle) -> Optional[Scenario]:
        scenarios = cls._get_all_by_cycle(cycle)
        for scenario in scenarios:
            if scenario.is_primary:
                return scenario
        return None

    @classmethod
    def _get_by_tag(cls, cycle: Cycle, tag: str) -> Optional[Scenario]:
        scenarios = cls._get_all_by_cycle(cycle)
        for scenario in scenarios:
            if scenario.has_tag(tag):
                return scenario
        return None

    @classmethod
    def _get_all_by_tag(cls, tag: str) -> List[Scenario]:
        scenarios = []
        for scenario in cls._get_all():
            if scenario.has_tag(tag):
                scenarios.append(scenario)
        return scenarios

    @classmethod
    def _get_all_by_cycle(cls, cycle: Cycle) -> List[Scenario]:
        filters = cls._build_filters_with_version("all")

        if not filters:
            filters = [{}]
        for fil in filters:
            fil.update({"cycle": cycle.id})
        return cls._get_all_by(filters)

    @classmethod
    def _get_primary_scenarios(cls) -> List[Scenario]:
        primary_scenarios = []
        for scenario in cls._get_all():
            if scenario.is_primary:
                primary_scenarios.append(scenario)
        return primary_scenarios

    @classmethod
    def _is_promotable_to_primary(cls, scenario: Union[Scenario, ScenarioId]) -> bool:
        if isinstance(scenario, str):
            scenario = cls._get(scenario)
        if scenario and not scenario.is_primary and scenario.cycle:
            return True
        return False

    @classmethod
    def _set_primary(cls, scenario: Scenario):
        if scenario.cycle:
            primary_scenario = cls._get_primary(scenario.cycle)
            # To prevent SAME scenario updating out of Context Manager
            if primary_scenario and primary_scenario != scenario:
                primary_scenario.is_primary = False  # type: ignore
            scenario.is_primary = True  # type: ignore
        else:
            raise DoesNotBelongToACycle(
                f"Can't set scenario {scenario.id} to primary because it doesn't belong to a cycle."
            )

    @classmethod
    def _tag(cls, scenario: Scenario, tag: str):
        tags = scenario.properties.get(cls._AUTHORIZED_TAGS_KEY, set())
        if len(tags) > 0 and tag not in tags:
            raise UnauthorizedTagError(f"Tag `{tag}` not authorized by scenario configuration `{scenario.config_id}`")
        if scenario.cycle:
            old_tagged_scenario = cls._get_by_tag(scenario.cycle, tag)
            if old_tagged_scenario:
                old_tagged_scenario.remove_tag(tag)
                cls._set(old_tagged_scenario)
        scenario._add_tag(tag)
        cls._set(scenario)
        _publish_event(cls._EVENT_ENTITY_TYPE, scenario.id, EventOperation.UPDATE, "tags")

    @classmethod
    def _untag(cls, scenario: Scenario, tag: str):
        scenario._remove_tag(tag)
        cls._set(scenario)
        _publish_event(cls._EVENT_ENTITY_TYPE, scenario.id, EventOperation.UPDATE, "tags")

    @classmethod
    def _compare(cls, *scenarios: Scenario, data_node_config_id: Optional[str] = None):
        if len(scenarios) < 2:
            raise InsufficientScenarioToCompare("At least two scenarios are required to compare.")

        if not all(scenarios[0].config_id == scenario.config_id for scenario in scenarios):
            raise DifferentScenarioConfigs("Scenarios to compare must have the same configuration.")

        if scenario_config := _ScenarioManager.__get_config(scenarios[0]):
            results = {}
            if data_node_config_id:
                if data_node_config_id in scenario_config.comparators.keys():
                    dn_comparators = {data_node_config_id: scenario_config.comparators[data_node_config_id]}
                else:
                    raise NonExistingComparator(f"Data node config {data_node_config_id} has no comparator.")
            else:
                dn_comparators = scenario_config.comparators

            for data_node_config_id, comparators in dn_comparators.items():
                data_nodes = [scenario.__getattr__(data_node_config_id).read() for scenario in scenarios]
                results[data_node_config_id] = {
                    comparator.__name__: comparator(*data_nodes) for comparator in comparators
                }

            return results

        else:
            raise NonExistingScenarioConfig(scenarios[0].config_id)

    @staticmethod
    def __get_config(scenario: Scenario):
        return Config.scenarios.get(scenario.config_id, None)

    @classmethod
    def _is_deletable(cls, scenario: Union[Scenario, ScenarioId]) -> bool:
        if isinstance(scenario, str):
            scenario = cls._get(scenario)
        if scenario.is_primary:
            if len(cls._get_all_by_cycle(scenario.cycle)) > 1:
                return False
        return True

    @classmethod
    def _delete(cls, scenario_id: ScenarioId):
        scenario = cls._get(scenario_id)
        if not cls._is_deletable(scenario):
            raise DeletingPrimaryScenario(
                f"Scenario {scenario.id}, which has config id {scenario.config_id}, is primary and there are "
                f"other scenarios in the same cycle. "
            )
        if scenario.is_primary:
            _CycleManagerFactory._build_manager()._delete(scenario.cycle.id)
        super()._delete(scenario_id)

    @classmethod
    def _hard_delete(cls, scenario_id: ScenarioId):
        scenario = cls._get(scenario_id)
        if not cls._is_deletable(scenario):
            raise DeletingPrimaryScenario(
                f"Scenario {scenario.id}, which has config id {scenario.config_id}, is primary and there are "
                f"other scenarios in the same cycle. "
            )
        if scenario.is_primary:
            _CycleManagerFactory._build_manager()._hard_delete(scenario.cycle.id)
        else:
            entity_ids_to_delete = cls._get_children_entity_ids(scenario)
            entity_ids_to_delete.scenario_ids.add(scenario.id)
            cls._delete_entities_of_multiple_types(entity_ids_to_delete)

    @classmethod
    def _delete_by_version(cls, version_number: str):
        """
        Deletes scenario by the version number.

        Check if the cycle is only attached to this scenario, then delete it.
        """
        for scenario in cls._repository._search("version", version_number):
            if scenario.cycle and len(cls._get_all_by_cycle(scenario.cycle)) == 1:
                _CycleManagerFactory._build_manager()._delete(scenario.cycle.id)
            super()._delete(scenario.id)

    @classmethod
    def _get_children_entity_ids(cls, scenario: Scenario) -> _EntityIds:
        entity_ids = _EntityIds()
        for sequence in scenario.sequences.values():
            if sequence.owner_id == scenario.id:
                entity_ids.sequence_ids.add(sequence.id)
        for task in scenario.tasks.values():
            if task.owner_id == scenario.id:
                entity_ids.task_ids.add(task.id)
        for data_node in scenario.data_nodes.values():
            if data_node.owner_id == scenario.id:
                entity_ids.data_node_ids.add(data_node.id)
        jobs = _JobManagerFactory._build_manager()._get_all()
        for job in jobs:
            if job.task.id in entity_ids.task_ids:
                entity_ids.job_ids.add(job.id)
        return entity_ids

    @classmethod
    def _get_by_config_id(cls, config_id: str, version_number: Optional[str] = None) -> List[Scenario]:
        """
        Get all scenarios by its config id.
        """
        filters = cls._build_filters_with_version(version_number)
        if not filters:
            filters = [{}]
        for fil in filters:
            fil.update({"config_id": config_id})
        return cls._repository._load_all(filters)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from copy import copy
from datetime import datetime, timedelta
from pydoc import locate

from .._repository._abstract_converter import _AbstractConverter
from .._version._utils import _migrate_entity
from ..common._utils import _load_fct
from ..data._data_model import _DataNodeModel
from ..data.data_node import DataNode
from . import GenericDataNode, JSONDataNode, MongoCollectionDataNode, SQLDataNode


class _DataNodeConverter(_AbstractConverter):
    _READ_FCT_NAME_KEY = "read_fct_name"
    _READ_FCT_MODULE_KEY = "read_fct_module"
    _WRITE_FCT_NAME_KEY = "write_fct_name"
    _WRITE_FCT_MODULE_KEY = "write_fct_module"
    _JSON_ENCODER_NAME_KEY = "encoder_name"
    _JSON_ENCODER_MODULE_KEY = "encoder_module"
    _JSON_DECODER_NAME_KEY = "decoder_name"
    _JSON_DECODER_MODULE_KEY = "decoder_module"
    _EXPOSED_TYPE_KEY = "exposed_type"
    __WRITE_QUERY_BUILDER_NAME_KEY = "write_query_builder_name"
    __WRITE_QUERY_BUILDER_MODULE_KEY = "write_query_builder_module"
    # TODO: This limits the valid string to only the ones provided by the Converter.
    # While in practice, each data nodes might have different exposed type possibilities.
    # The previous implementation used tabular datanode but it's no longer suitable so
    # new proposal is needed.
    _VALID_STRING_EXPOSED_TYPES = ["numpy", "pandas", "modin"]

    @classmethod
    def __serialize_generic_dn_properties(cls, datanode_properties: dict):
        read_fct = datanode_properties.get(GenericDataNode._OPTIONAL_READ_FUNCTION_PROPERTY, None)
        datanode_properties[cls._READ_FCT_NAME_KEY] = read_fct.__name__ if read_fct else None
        datanode_properties[cls._READ_FCT_MODULE_KEY] = read_fct.__module__ if read_fct else None

        write_fct = datanode_properties.get(GenericDataNode._OPTIONAL_WRITE_FUNCTION_PROPERTY, None)
        datanode_properties[cls._WRITE_FCT_NAME_KEY] = write_fct.__name__ if write_fct else None
        datanode_properties[cls._WRITE_FCT_MODULE_KEY] = write_fct.__module__ if write_fct else None

        del (
            datanode_properties[GenericDataNode._OPTIONAL_READ_FUNCTION_PROPERTY],
            datanode_properties[GenericDataNode._OPTIONAL_WRITE_FUNCTION_PROPERTY],
        )
        return datanode_properties

    @classmethod
    def __serialize_json_dn_properties(cls, datanode_properties: dict):
        encoder = datanode_properties.get(JSONDataNode._ENCODER_KEY)
        datanode_properties[cls._JSON_ENCODER_NAME_KEY] = encoder.__name__ if encoder else None
        datanode_properties[cls._JSON_ENCODER_MODULE_KEY] = encoder.__module__ if encoder else None
        datanode_properties.pop(JSONDataNode._ENCODER_KEY, None)

        decoder = datanode_properties.get(JSONDataNode._DECODER_KEY)
        datanode_properties[cls._JSON_DECODER_NAME_KEY] = decoder.__name__ if decoder else None
        datanode_properties[cls._JSON_DECODER_MODULE_KEY] = decoder.__module__ if decoder else None
        datanode_properties.pop(JSONDataNode._DECODER_KEY, None)

        return datanode_properties

    @classmethod
    def __serialize_sql_dn_properties(cls, datanode_properties: dict) -> dict:
        query_builder = datanode_properties.get(SQLDataNode._WRITE_QUERY_BUILDER_KEY)
        datanode_properties[cls.__WRITE_QUERY_BUILDER_NAME_KEY] = query_builder.__name__ if query_builder else None
        datanode_properties[cls.__WRITE_QUERY_BUILDER_MODULE_KEY] = query_builder.__module__ if query_builder else None
        datanode_properties.pop(SQLDataNode._WRITE_QUERY_BUILDER_KEY, None)

        return datanode_properties

    @classmethod
    def __serialize_mongo_collection_dn_model_properties(cls, datanode_properties: dict) -> dict:
        if MongoCollectionDataNode._CUSTOM_DOCUMENT_PROPERTY in datanode_properties.keys():
            datanode_properties[MongoCollectionDataNode._CUSTOM_DOCUMENT_PROPERTY] = (
                f"{datanode_properties[MongoCollectionDataNode._CUSTOM_DOCUMENT_PROPERTY].__module__}."
                f"{datanode_properties[MongoCollectionDataNode._CUSTOM_DOCUMENT_PROPERTY].__qualname__}"
            )

        return datanode_properties

    @classmethod
    def __serialize_edits(cls, edits):
        new_edits = []
        for edit in edits:
            new_edit = edit.copy()
            if timestamp := new_edit.get("timestamp", None):
                new_edit["timestamp"] = timestamp.isoformat()
            else:
                new_edit["timestamp"] = datetime.now().isoformat()
            new_edits.append(new_edit)
        return new_edits

    @staticmethod
    def __serialize_exposed_type(properties: dict, exposed_type_key: str, valid_str_exposed_types) -> dict:
        if not isinstance(properties[exposed_type_key], str):
            if isinstance(properties[exposed_type_key], dict):
                properties[exposed_type_key] = {
                    k: v if v in valid_str_exposed_types else f"{v.__module__}.{v.__qualname__}"
                    for k, v in properties[exposed_type_key].items()
                }
            elif isinstance(properties[exposed_type_key], list):
                properties[exposed_type_key] = [
                    v if v in valid_str_exposed_types else f"{v.__module__}.{v.__qualname__}"
                    for v in properties[exposed_type_key]
                ]
            else:
                properties[
                    exposed_type_key
                ] = f"{properties[exposed_type_key].__module__}.{properties[exposed_type_key].__qualname__}"
        return properties

    @classmethod
    def _entity_to_model(cls, data_node: DataNode) -> _DataNodeModel:
        properties = data_node._properties.data.copy()
        if data_node.storage_type() == GenericDataNode.storage_type():
            properties = cls.__serialize_generic_dn_properties(properties)

        if data_node.storage_type() == JSONDataNode.storage_type():
            properties = cls.__serialize_json_dn_properties(properties)

        if data_node.storage_type() == SQLDataNode.storage_type():
            properties = cls.__serialize_sql_dn_properties(properties)

        if data_node.storage_type() == MongoCollectionDataNode.storage_type():
            properties = cls.__serialize_mongo_collection_dn_model_properties(properties)

        if cls._EXPOSED_TYPE_KEY in properties.keys():
            properties = cls.__serialize_exposed_type(
                properties, cls._EXPOSED_TYPE_KEY, cls._VALID_STRING_EXPOSED_TYPES
            )

        return _DataNodeModel(
            data_node.id,
            data_node.config_id,
            data_node._scope,
            data_node.storage_type(),
            properties.pop("name", None) or data_node._name,
            data_node.owner_id,
            list(data_node._parent_ids),
            data_node._last_edit_date.isoformat() if data_node._last_edit_date else None,
            cls.__serialize_edits(data_node._edits),
            data_node._version,
            data_node._validity_period.days if data_node._validity_period else None,
            data_node._validity_period.seconds if data_node._validity_period else None,
            data_node._edit_in_progress,
            data_node._editor_id,
            data_node._editor_expiration_date.isoformat() if data_node._editor_expiration_date else None,
            properties,
        )

    @classmethod
    def __deserialize_generic_dn_properties(cls, datanode_model_properties):
        if datanode_model_properties[cls._READ_FCT_MODULE_KEY]:
            datanode_model_properties[GenericDataNode._OPTIONAL_READ_FUNCTION_PROPERTY] = _load_fct(
                datanode_model_properties[cls._READ_FCT_MODULE_KEY],
                datanode_model_properties[cls._READ_FCT_NAME_KEY],
            )
        else:
            datanode_model_properties[GenericDataNode._OPTIONAL_READ_FUNCTION_PROPERTY] = None

        if datanode_model_properties[cls._WRITE_FCT_MODULE_KEY]:
            datanode_model_properties[GenericDataNode._OPTIONAL_WRITE_FUNCTION_PROPERTY] = _load_fct(
                datanode_model_properties[cls._WRITE_FCT_MODULE_KEY],
                datanode_model_properties[cls._WRITE_FCT_NAME_KEY],
            )
        else:
            datanode_model_properties[GenericDataNode._OPTIONAL_WRITE_FUNCTION_PROPERTY] = None

        del datanode_model_properties[cls._READ_FCT_NAME_KEY]
        del datanode_model_properties[cls._READ_FCT_MODULE_KEY]
        del datanode_model_properties[cls._WRITE_FCT_NAME_KEY]
        del datanode_model_properties[cls._WRITE_FCT_MODULE_KEY]

        return datanode_model_properties

    @classmethod
    def __deserialize_json_dn_properties(cls, datanode_model_properties: dict) -> dict:
        if datanode_model_properties[cls._JSON_ENCODER_MODULE_KEY]:
            datanode_model_properties[JSONDataNode._ENCODER_KEY] = _load_fct(
                datanode_model_properties[cls._JSON_ENCODER_MODULE_KEY],
                datanode_model_properties[cls._JSON_ENCODER_NAME_KEY],
            )
        else:
            datanode_model_properties[JSONDataNode._ENCODER_KEY] = None

        if datanode_model_properties[cls._JSON_DECODER_MODULE_KEY]:
            datanode_model_properties[JSONDataNode._DECODER_KEY] = _load_fct(
                datanode_model_properties[cls._JSON_DECODER_MODULE_KEY],
                datanode_model_properties[cls._JSON_DECODER_NAME_KEY],
            )
        else:
            datanode_model_properties[JSONDataNode._DECODER_KEY] = None

        del datanode_model_properties[cls._JSON_ENCODER_NAME_KEY]
        del datanode_model_properties[cls._JSON_ENCODER_MODULE_KEY]
        del datanode_model_properties[cls._JSON_DECODER_NAME_KEY]
        del datanode_model_properties[cls._JSON_DECODER_MODULE_KEY]

        return datanode_model_properties

    @classmethod
    def __deserialize_sql_dn_model_properties(cls, datanode_model_properties: dict) -> dict:

        if datanode_model_properties[cls.__WRITE_QUERY_BUILDER_MODULE_KEY]:
            datanode_model_properties[SQLDataNode._WRITE_QUERY_BUILDER_KEY] = _load_fct(
                datanode_model_properties[cls.__WRITE_QUERY_BUILDER_MODULE_KEY],
                datanode_model_properties[cls.__WRITE_QUERY_BUILDER_NAME_KEY],
            )
        else:
            datanode_model_properties[SQLDataNode._WRITE_QUERY_BUILDER_KEY] = None

        del datanode_model_properties[cls.__WRITE_QUERY_BUILDER_NAME_KEY]
        del datanode_model_properties[cls.__WRITE_QUERY_BUILDER_MODULE_KEY]

        return datanode_model_properties

    @classmethod
    def __deserialize_mongo_collection_dn_model_properties(cls, datanode_model_properties: dict) -> dict:
        if MongoCollectionDataNode._CUSTOM_DOCUMENT_PROPERTY in datanode_model_properties.keys():
            if isinstance(datanode_model_properties[MongoCollectionDataNode._CUSTOM_DOCUMENT_PROPERTY], str):
                datanode_model_properties[MongoCollectionDataNode._CUSTOM_DOCUMENT_PROPERTY] = locate(
                    datanode_model_properties[MongoCollectionDataNode._CUSTOM_DOCUMENT_PROPERTY]
                )
        return datanode_model_properties

    @classmethod
    def __deserialize_edits(cls, edits):
        for edit in edits:
            if timestamp := edit.get("timestamp", None):
                edit["timestamp"] = datetime.fromisoformat(timestamp)
            else:
                edit["timestamp"] = datetime.now()
        return edits

    @staticmethod
    def __deserialize_exposed_type(properties: dict, exposed_type_key: str, valid_str_exposed_types) -> dict:
        if properties[exposed_type_key] not in valid_str_exposed_types:
            if isinstance(properties[exposed_type_key], str):
                properties[exposed_type_key] = locate(properties[exposed_type_key])
            elif isinstance(properties[exposed_type_key], dict):
                properties[exposed_type_key] = {
                    k: v if v in valid_str_exposed_types else locate(v) for k, v in properties[exposed_type_key].items()
                }
            elif isinstance(properties[exposed_type_key], list):
                properties[exposed_type_key] = [
                    v if v in valid_str_exposed_types else locate(v) for v in properties[exposed_type_key]
                ]
        return properties

    @classmethod
    def _model_to_entity(cls, model: _DataNodeModel) -> DataNode:
        data_node_properties = model.data_node_properties.copy()

        if model.storage_type == GenericDataNode.storage_type():
            data_node_properties = cls.__deserialize_generic_dn_properties(data_node_properties)

        if model.storage_type == JSONDataNode.storage_type():
            data_node_properties = cls.__deserialize_json_dn_properties(data_node_properties)

        if model.storage_type == SQLDataNode.storage_type():
            data_node_properties = cls.__deserialize_sql_dn_model_properties(data_node_properties)

        if model.storage_type == MongoCollectionDataNode.storage_type():
            data_node_properties = cls.__deserialize_mongo_collection_dn_model_properties(data_node_properties)

        if cls._EXPOSED_TYPE_KEY in data_node_properties.keys():
            data_node_properties = cls.__deserialize_exposed_type(
                data_node_properties, cls._EXPOSED_TYPE_KEY, cls._VALID_STRING_EXPOSED_TYPES
            )

        validity_period = None
        if model.validity_seconds is not None and model.validity_days is not None:
            validity_period = timedelta(days=model.validity_days, seconds=model.validity_seconds)

        exp_date = datetime.fromisoformat(model.editor_expiration_date) if model.editor_expiration_date else None
        datanode = DataNode._class_map()[model.storage_type](
            config_id=model.config_id,
            scope=model.scope,
            id=model.id,
            name=model.name,
            owner_id=model.owner_id,
            parent_ids=set(model.parent_ids),
            last_edit_date=datetime.fromisoformat(model.last_edit_date) if model.last_edit_date else None,
            edits=cls.__deserialize_edits(copy(model.edits)),
            version=model.version,
            validity_period=validity_period,
            edit_in_progress=model.edit_in_progress,
            editor_id=model.editor_id,
            editor_expiration_date=exp_date,
            properties=data_node_properties,
        )
        return _migrate_entity(datanode)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import json
import os
import shutil
from typing import Dict

from taipy.logger._taipy_logger import _TaipyLogger

from ._utils import _migrate

__logger = _TaipyLogger._get_logger()


def _load_all_entities_from_fs(root: str) -> Dict:
    # run through all files in the data folder and load them
    entities = {}
    for root, dirs, files in os.walk(root):
        for file in files:
            if file.endswith(".json"):
                with open(os.path.join(root, file)) as f:
                    _id = file.split(".")[0]
                    if "version" in root:
                        _id = f"VERSION_{_id}"
                    entities[_id] = {
                        "data": json.load(f),
                        "path": os.path.join(root, file),
                    }
    return entities


def __write_entities_to_fs(_entities: Dict, root: str):
    if not os.path.exists(root):
        os.makedirs(root, exist_ok=True)

    for _id, entity in _entities.items():
        # Do not write pipeline entities
        if "PIPELINE" in _id:
            continue
        with open(entity["path"], "w") as f:
            json.dump(entity["data"], f, indent=0)

    # Remove pipelines folder
    pipelines_path = os.path.join(root, "pipelines")
    if os.path.exists(pipelines_path):
        shutil.rmtree(pipelines_path)


def _restore_migrate_file_entities(path: str) -> bool:
    backup_path = f"{path}_backup"

    if not os.path.exists(backup_path):
        __logger.error(f"The backup folder '{backup_path}' does not exist.")
        return False

    if os.path.exists(path):
        shutil.rmtree(path)
    else:
        __logger.warning(f"The original entities folder '{path}' does not exist.")

    os.rename(backup_path, path)
    __logger.info(f"Restored entities from the backup folder '{backup_path}' to '{path}'.")
    return True


def _remove_backup_file_entities(path: str) -> bool:
    backup_path = f"{path}_backup"
    if not os.path.exists(backup_path):
        __logger.error(f"The backup folder '{backup_path}' does not exist.")
        return False

    shutil.rmtree(backup_path)
    __logger.info(f"Removed backup entities from the backup folder '{backup_path}'.")
    return True


def _migrate_fs_entities(path: str, backup: bool = True) -> bool:
    """Migrate entities from filesystem to the current version.

    Args:
        path (str): The path to the folder containing the entities.
        backup (bool, optional): Whether to backup the entities before migrating. Defaults to True.

    Returns:
        bool: True if the migration was successful, False otherwise.
    """
    if not os.path.isdir(path):
        __logger.error(f"Folder '{path}' does not exist.")
        return False

    if backup:
        backup_path = f"{path}_backup"
        try:
            shutil.copytree(path, backup_path)
        except FileExistsError:
            __logger.warning(f"The backup folder '{backup_path}' already exists. Migration canceled.")
            return False
        else:
            __logger.info(f"Backed up entities from '{path}' to '{backup_path}' folder before migration.")

    __logger.info(f"Starting entity migration from '{path}' folder.")

    entities = _load_all_entities_from_fs(path)
    entities, _ = _migrate(entities)
    __write_entities_to_fs(entities, path)

    __logger.info("Migration finished")
    return True



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy.config._config import _Config
from taipy.config.checker._checkers._config_checker import _ConfigChecker
from taipy.config.checker.issue_collector import IssueCollector

from ..._version._version_manager_factory import _VersionManagerFactory
from ..migration_config import MigrationConfig


class _MigrationConfigChecker(_ConfigChecker):
    def __init__(self, config: _Config, collector: IssueCollector):
        super().__init__(config, collector)

    def _check(self) -> IssueCollector:
        if migration_config := self._config._unique_sections.get(MigrationConfig.name):
            self._check_if_entity_property_key_used_is_predefined(migration_config)

            migration_fcts = migration_config.migration_fcts

            for target_version, migration_functions in migration_config.migration_fcts.items():
                for config_id, migration_function in migration_functions.items():
                    self._check_callable(target_version, config_id, migration_function)

            self._check_valid_production_version(migration_fcts)
            self._check_migration_from_productions_to_productions_exist(migration_fcts)

        return self._collector

    def _check_callable(self, target_version, config_id, migration_function):
        if not callable(migration_function):
            self._error(
                MigrationConfig._MIGRATION_FCTS_KEY,
                migration_function,
                f"The migration function of config `{config_id}` from version {target_version}"
                f" must be populated with Callable value.",
            )

    def _check_valid_production_version(self, migration_fcts):
        for target_version in migration_fcts.keys():
            if target_version not in _VersionManagerFactory._build_manager()._get_production_versions():
                self._error(
                    MigrationConfig._MIGRATION_FCTS_KEY,
                    target_version,
                    "The target version for a migration function must be a production version.",
                )

    def _check_migration_from_productions_to_productions_exist(self, migration_fcts):
        production_versions = _VersionManagerFactory._build_manager()._get_production_versions()
        for source_version, target_version in zip(production_versions[:-1], production_versions[1:]):
            if not migration_fcts.get(target_version):
                self._info(
                    "target_version",
                    None,
                    f'There is no migration function from production version "{source_version}"'
                    f' to version "{target_version}".',
                )



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Type

from .._manager._manager_factory import _ManagerFactory
from ..common._utils import _load_fct
from ._scenario_fs_repository import _ScenarioFSRepository
from ._scenario_manager import _ScenarioManager
from ._scenario_sql_repository import _ScenarioSQLRepository


class _ScenarioManagerFactory(_ManagerFactory):

    __REPOSITORY_MAP = {"default": _ScenarioFSRepository, "sql": _ScenarioSQLRepository}

    @classmethod
    def _build_manager(cls) -> Type[_ScenarioManager]:  # type: ignore
        if cls._using_enterprise():
            scenario_manager = _load_fct(
                cls._TAIPY_ENTERPRISE_CORE_MODULE + ".scenario._scenario_manager", "_ScenarioManager"
            )  # type: ignore
            build_repository = _load_fct(
                cls._TAIPY_ENTERPRISE_CORE_MODULE + ".scenario._scenario_manager_factory", "_ScenarioManagerFactory"
            )._build_repository  # type: ignore
        else:
            scenario_manager = _ScenarioManager
            build_repository = cls._build_repository
        scenario_manager._repository = build_repository()  # type: ignore
        return scenario_manager  # type: ignore

    @classmethod
    def _build_repository(cls):
        return cls._get_repository_with_repo_map(cls.__REPOSITORY_MAP)()



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import json
import pathlib
from typing import Any, Dict, Iterable, List, Optional, Type, Union

from sqlalchemy.exc import NoResultFound

from ..common.typing import Converter, Entity, ModelType
from ..exceptions import ModelNotFound
from ._abstract_repository import _AbstractRepository
from .db._sql_session import _SQLSession


class _SQLRepository(_AbstractRepository[ModelType, Entity]):
    def __init__(self, model_type: Type[ModelType], converter: Type[Converter], session=None):
        """
        Holds common methods to be used and extended when the need for saving
        dataclasses in a SqlLite database.

        Some lines have type: ignore because MyPy won't recognize some generic attributes. This
        should be revised in the future.

        Attributes:
            model_type: Generic dataclass.
            converter: A class that handles conversion to and from a database backend
            db: An SQLAlchemy session object
        """
        SessionLocal = _SQLSession.init_db()
        self.db = session or SessionLocal()
        self.model_type = model_type
        self.converter = converter

    ###############################
    # ##   Inherited methods   ## #
    ###############################
    def _save(self, entity: Entity):
        obj = self.converter._entity_to_model(entity)
        if self.db.query(self.model_type).filter_by(id=obj.id).first():
            self.__update_entry(obj)
            return
        self.__insert_model(obj)

    def _exists(self, entity_id: str):
        return bool(self.db.query(self.model_type.id).filter_by(id=entity_id).first())  # type: ignore

    def _load(self, entity_id: str) -> Entity:
        if entry := self.db.query(self.model_type).filter(self.model_type.id == entity_id).first():  # type: ignore
            return self.converter._model_to_entity(entry)
        raise ModelNotFound(str(self.model_type.__name__), entity_id)

    def _load_all(self, filters: Optional[List[Dict]] = None) -> List[Entity]:
        query = self.db.query(self.model_type)
        entities: List[Entity] = []

        for f in filters or [{}]:
            filtered_query = query.filter_by(**f)
            try:
                entities.extend([self.converter._model_to_entity(m) for m in filtered_query.all()])
            except NoResultFound:
                continue
        return entities

    def _delete(self, entity_id: str):
        number_of_deleted_entries = self.db.query(self.model_type).filter_by(id=entity_id).delete()
        if not number_of_deleted_entries:
            raise ModelNotFound(str(self.model_type.__name__), entity_id)
        self.db.commit()

    def _delete_all(self):
        self.db.query(self.model_type).delete()
        self.db.commit()

    def _delete_many(self, ids: Iterable[str]):
        for entity_id in ids:
            self._delete(entity_id)

    def _delete_by(self, attribute: str, value: str):
        self.db.query(self.model_type).filter_by(**{attribute: value}).delete()
        self.db.commit()

    def _search(self, attribute: str, value: Any, filters: Optional[List[Dict]] = None) -> List[Entity]:
        query = self.db.query(self.model_type).filter_by(**{attribute: value})

        entities: List[Entity] = []
        for f in filters or [{}]:
            filtered_query = query.filter_by(**f)
            entities.extend([self.converter._model_to_entity(m) for m in filtered_query.all()])

        return entities

    def _export(self, entity_id: str, folder_path: Union[str, pathlib.Path]):
        if isinstance(folder_path, str):
            folder: pathlib.Path = pathlib.Path(folder_path)
        else:
            folder = folder_path

        export_dir = folder / self.model_type.__table__.name  # type: ignore
        if not export_dir.exists():
            export_dir.mkdir(parents=True)

        export_path = export_dir / f"{entity_id}.json"

        entry = self.db.query(self.model_type).filter_by(id=entity_id).first()
        if entry is None:
            raise ModelNotFound(self.model_type, entity_id)  # type: ignore

        with open(export_path, "w", encoding="utf-8") as export_file:
            export_file.write(json.dumps(entry.to_dict()))

    ###########################################
    # ##   Specific or optimized methods   ## #
    ###########################################
    def _get_multi(self, *, skip: int = 0, limit: int = 100) -> List[ModelType]:
        return self.db.query(self.model_type).offset(skip).limit(limit).all()

    def _get_by_config(self, config_id: Any) -> Optional[ModelType]:
        return self.db.query(self.model_type).filter(self.model_type.config_id == config_id).first()  # type: ignore

    def _get_by_config_and_owner_id(
        self, config_id: str, owner_id: Optional[str], filters: Optional[List[Dict]] = None
    ) -> Optional[Entity]:
        if not filters:
            filters = [{}]
        if entry := self.__get_entities_by_config_and_owner(config_id, owner_id, filters):
            return self.converter._model_to_entity(entry)
        return None

    def _get_by_configs_and_owner_ids(self, configs_and_owner_ids, filters: Optional[List[Dict]] = None):
        # Design in order to optimize performance on Entity creation.
        # Maintainability and readability were impacted.
        if not filters:
            filters = [{}]
        res = {}
        configs_and_owner_ids = set(configs_and_owner_ids)

        for config, owner in configs_and_owner_ids:
            entry = self.__get_entities_by_config_and_owner(config.id, owner, filters)
            if entry:
                entity = self.converter._model_to_entity(entry)
                key = config, owner
                res[key] = entity

        return res

    def __get_entities_by_config_and_owner(
        self, config_id: str, owner_id: Optional[str] = "", filters: Optional[List[Dict]] = None
    ) -> ModelType:
        if not filters:
            filters = []
        versions = [item.get("version") for item in filters if item.get("version")]
        if owner_id:
            query = self.db.query(self.model_type).filter_by(config_id=config_id).filter_by(owner_id=owner_id)
        else:
            query = self.db.query(self.model_type).filter_by(config_id=config_id).filter_by(owner_id=None)
        if versions:
            query = query.filter(self.model_type.version.in_(versions))  # type: ignore
        return query.first()

    #############################
    # ##   Private methods   ## #
    #############################
    def __insert_model(self, model: ModelType):
        self.db.add(model)
        self.db.commit()
        self.db.refresh(model)

    def __update_entry(self, model):
        self.db.merge(model)
        self.db.commit()



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from sqlalchemy import JSON, Boolean, Column, String, Table

from .._repository._base_taipy_model import _BaseModel
from .._repository.db._sql_base_model import mapper_registry


@mapper_registry.mapped
@dataclass
class _TaskModel(_BaseModel):
    __table__ = Table(
        "task",
        mapper_registry.metadata,
        Column("id", String, primary_key=True),
        Column("owner_id", String),
        Column("parent_ids", JSON),
        Column("config_id", String),
        Column("input_ids", JSON),
        Column("function_name", String),
        Column("function_module", String),
        Column("output_ids", JSON),
        Column("version", String),
        Column("skippable", Boolean),
        Column("properties", JSON),
    )
    id: str
    owner_id: Optional[str]
    parent_ids: List[str]
    config_id: str
    input_ids: List[str]
    function_name: str
    function_module: str
    output_ids: List[str]
    version: str
    skippable: bool
    properties: Dict[str, Any]

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        return _TaskModel(
            id=data["id"],
            owner_id=data.get("owner_id"),
            parent_ids=data.get("parent_ids", []),
            config_id=data["config_id"],
            input_ids=data["input_ids"],
            function_name=data["function_name"],
            function_module=data["function_module"],
            output_ids=data["output_ids"],
            version=data["version"],
            skippable=data["skippable"],
            properties=data["properties"] if "properties" in data.keys() else {},
        )



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import datetime
import os
import pathlib
import shutil
from unittest import mock

import pytest

import src.taipy.core.taipy as tp
from src.taipy.core import (
    Core,
    Cycle,
    CycleId,
    DataNodeId,
    JobId,
    Scenario,
    ScenarioId,
    Sequence,
    SequenceId,
    Task,
    TaskId,
)
from src.taipy.core._orchestrator._orchestrator_factory import _OrchestratorFactory
from src.taipy.core._version._version_manager import _VersionManager
from src.taipy.core.config.data_node_config import DataNodeConfig
from src.taipy.core.config.job_config import JobConfig
from src.taipy.core.config.scenario_config import ScenarioConfig
from src.taipy.core.cycle._cycle_manager import _CycleManager
from src.taipy.core.data._data_manager import _DataManager
from src.taipy.core.data.pickle import PickleDataNode
from src.taipy.core.exceptions.exceptions import DataNodeConfigIsNotGlobal, InvalidExportPath
from src.taipy.core.job._job_manager import _JobManager
from src.taipy.core.job.job import Job
from src.taipy.core.scenario._scenario_manager import _ScenarioManager
from src.taipy.core.sequence._sequence_manager import _SequenceManager
from src.taipy.core.task._task_manager import _TaskManager
from taipy.config.common.frequency import Frequency
from taipy.config.common.scope import Scope
from taipy.config.config import Config
from taipy.config.exceptions.exceptions import ConfigurationUpdateBlocked


class TestTaipy:
    def test_set(self, scenario, cycle, sequence, data_node, task):
        with mock.patch("src.taipy.core.data._data_manager._DataManager._set") as mck:
            tp.set(data_node)
            mck.assert_called_once_with(data_node)
        with mock.patch("src.taipy.core.task._task_manager._TaskManager._set") as mck:
            tp.set(task)
            mck.assert_called_once_with(task)
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._set") as mck:
            tp.set(sequence)
            mck.assert_called_once_with(sequence)
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._set") as mck:
            tp.set(scenario)
            mck.assert_called_once_with(scenario)
        with mock.patch("src.taipy.core.cycle._cycle_manager._CycleManager._set") as mck:
            tp.set(cycle)
            mck.assert_called_once_with(cycle)

    def test_is_editable_is_called(self, cycle, job, data_node):
        with mock.patch("src.taipy.core.cycle._cycle_manager._CycleManager._is_editable") as mck:
            cycle_id = CycleId("CYCLE_id")
            tp.is_editable(cycle_id)
            mck.assert_called_once_with(cycle_id)

        with mock.patch("src.taipy.core.cycle._cycle_manager._CycleManager._is_editable") as mck:
            tp.is_editable(cycle)
            mck.assert_called_once_with(cycle)

        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_editable") as mck:
            scenario_id = ScenarioId("SCENARIO_id")
            tp.is_editable(scenario_id)
            mck.assert_called_once_with(scenario_id)

        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_editable") as mck:
            scenario = Scenario("scenario_config_id", [], {})
            tp.is_editable(scenario)
            mck.assert_called_once_with(scenario)

        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._is_editable") as mck:
            sequence_id = SequenceId("SEQUENCE_id")
            tp.is_editable(sequence_id)
            mck.assert_called_once_with(sequence_id)

        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._is_editable") as mck:
            sequence = Sequence({}, [], "sequence_id")
            tp.is_editable(sequence)
            mck.assert_called_once_with(sequence)

        with mock.patch("src.taipy.core.task._task_manager._TaskManager._is_editable") as mck:
            task_id = TaskId("TASK_id")
            tp.is_editable(task_id)
            mck.assert_called_once_with(task_id)

        with mock.patch("src.taipy.core.task._task_manager._TaskManager._is_editable") as mck:
            task = Task("task_config_id", {}, print)
            tp.is_editable(task)
            mck.assert_called_once_with(task)

        with mock.patch("src.taipy.core.job._job_manager._JobManager._is_editable") as mck:
            job_id = JobId("JOB_id")
            tp.is_editable(job_id)
            mck.assert_called_once_with(job_id)

        with mock.patch("src.taipy.core.job._job_manager._JobManager._is_editable") as mck:
            tp.is_editable(job)
            mck.assert_called_once_with(job)

        with mock.patch("src.taipy.core.data._data_manager._DataManager._is_editable") as mck:
            dn_id = DataNodeId("DATANODE_id")
            tp.is_editable(dn_id)
            mck.assert_called_once_with(dn_id)

        with mock.patch("src.taipy.core.data._data_manager._DataManager._is_editable") as mck:
            tp.is_editable(data_node)
            mck.assert_called_once_with(data_node)

    def test_is_editable(self):
        current_date = datetime.datetime.now()

        cycle = Cycle(Frequency.DAILY, {}, current_date, current_date, current_date)
        scenario = Scenario("scenario_config_id", [], {}, sequences={"sequence": {}})

        task = Task("task_config_id", {}, print)
        job = Job("job_id", task, "submit_id", scenario.id)
        dn = PickleDataNode("data_node_config_id", Scope.SCENARIO)

        _CycleManager._set(cycle)
        _ScenarioManager._set(scenario)
        _TaskManager._set(task)
        _JobManager._set(job)
        _DataManager._set(dn)
        sequence = scenario.sequences["sequence"]

        assert tp.is_editable(scenario)
        assert tp.is_editable(sequence)
        assert tp.is_editable(task)
        assert tp.is_editable(cycle)
        assert tp.is_editable(job)
        assert tp.is_editable(dn)

    def test_is_readable_is_called(self, cycle, job, data_node):
        with mock.patch("src.taipy.core.cycle._cycle_manager._CycleManager._is_readable") as mck:
            cycle_id = CycleId("CYCLE_id")
            tp.is_readable(cycle_id)
            mck.assert_called_once_with(cycle_id)

        with mock.patch("src.taipy.core.cycle._cycle_manager._CycleManager._is_readable") as mck:
            tp.is_readable(cycle)
            mck.assert_called_once_with(cycle)

        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_readable") as mck:
            scenario_id = ScenarioId("SCENARIO_id")
            tp.is_readable(scenario_id)
            mck.assert_called_once_with(scenario_id)

        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_readable") as mck:
            scenario = Scenario("scenario_config_id", [], {})
            tp.is_readable(scenario)
            mck.assert_called_once_with(scenario)

        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._is_readable") as mck:
            sequence_id = SequenceId("SEQUENCE_id")
            tp.is_readable(sequence_id)
            mck.assert_called_once_with(sequence_id)

        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._is_readable") as mck:
            sequence = Sequence({}, [], "sequence_id")
            tp.is_readable(sequence)
            mck.assert_called_once_with(sequence)

        with mock.patch("src.taipy.core.task._task_manager._TaskManager._is_readable") as mck:
            task_id = TaskId("TASK_id")
            tp.is_readable(task_id)
            mck.assert_called_once_with(task_id)

        with mock.patch("src.taipy.core.task._task_manager._TaskManager._is_readable") as mck:
            task = Task("task_config_id", {}, print)
            tp.is_readable(task)
            mck.assert_called_once_with(task)

        with mock.patch("src.taipy.core.job._job_manager._JobManager._is_readable") as mck:
            job_id = JobId("JOB_id")
            tp.is_readable(job_id)
            mck.assert_called_once_with(job_id)

        with mock.patch("src.taipy.core.job._job_manager._JobManager._is_readable") as mck:
            tp.is_readable(job)
            mck.assert_called_once_with(job)

        with mock.patch("src.taipy.core.data._data_manager._DataManager._is_readable") as mck:
            dn_id = DataNodeId("DATANODE_id")
            tp.is_readable(dn_id)
            mck.assert_called_once_with(dn_id)

        with mock.patch("src.taipy.core.data._data_manager._DataManager._is_readable") as mck:
            tp.is_readable(data_node)
            mck.assert_called_once_with(data_node)

    def test_is_readable(self):
        current_date = datetime.datetime.now()

        cycle = Cycle(Frequency.DAILY, {}, current_date, current_date, current_date)
        scenario = Scenario("scenario_config_id", [], {}, sequences={"sequence": {}})

        task = Task("task_config_id", {}, print)
        job = Job("job_id", task, "submit_id", scenario.id)
        dn = PickleDataNode("data_node_config_id", Scope.SCENARIO)

        _CycleManager._set(cycle)
        _ScenarioManager._set(scenario)
        _TaskManager._set(task)
        _JobManager._set(job)
        _DataManager._set(dn)
        sequence = scenario.sequences["sequence"]

        assert tp.is_readable(scenario)
        assert tp.is_readable(sequence)
        assert tp.is_readable(task)
        assert tp.is_readable(cycle)
        assert tp.is_readable(job)
        assert tp.is_readable(dn)

    def test_is_submittable_is_called(self):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_submittable") as mck:
            scenario_id = ScenarioId("SCENARIO_id")
            tp.is_submittable(scenario_id)
            mck.assert_called_once_with(scenario_id)

        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_submittable") as mck:
            scenario = Scenario("scenario_config_id", [], {})
            tp.is_submittable(scenario)
            mck.assert_called_once_with(scenario)

        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._is_submittable") as mck:
            sequence_id = SequenceId("SEQUENCE_id")
            tp.is_submittable(sequence_id)
            mck.assert_called_once_with(sequence_id)

        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._is_submittable") as mck:
            sequence = Sequence({}, [], "sequence_id")
            tp.is_submittable(sequence)
            mck.assert_called_once_with(sequence)

        with mock.patch("src.taipy.core.task._task_manager._TaskManager._is_submittable") as mck:
            task_id = TaskId("TASK_id")
            tp.is_submittable(task_id)
            mck.assert_called_once_with(task_id)

        with mock.patch("src.taipy.core.task._task_manager._TaskManager._is_submittable") as mck:
            task = Task("task_config_id", {}, print)
            tp.is_submittable(task)
            mck.assert_called_once_with(task)

    def test_is_submittable(self):
        current_date = datetime.datetime.now()

        cycle = Cycle(Frequency.DAILY, {}, current_date, current_date, current_date)
        scenario = Scenario("scenario_config_id", [], {}, sequences={"sequence": {}})

        task = Task("task_config_id", {}, print)
        job = Job("job_id", task, "submit_id", scenario.id)
        dn = PickleDataNode("data_node_config_id", Scope.SCENARIO)

        _CycleManager._set(cycle)
        _ScenarioManager._set(scenario)
        _TaskManager._set(task)
        _JobManager._set(job)
        _DataManager._set(dn)
        sequence = scenario.sequences["sequence"]

        assert tp.is_submittable(scenario)
        assert tp.is_submittable(sequence)
        assert tp.is_submittable(task)
        assert not tp.is_submittable(cycle)
        assert not tp.is_submittable(job)
        assert not tp.is_submittable(dn)

    def test_submit(self, scenario, sequence, task):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._submit") as mck:
            tp.submit(scenario)
            mck.assert_called_once_with(scenario, force=False, wait=False, timeout=None)
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._submit") as mck:
            tp.submit(sequence)
            mck.assert_called_once_with(sequence, force=False, wait=False, timeout=None)
        with mock.patch("src.taipy.core.task._task_manager._TaskManager._submit") as mck:
            tp.submit(task)
            mck.assert_called_once_with(task, force=False, wait=False, timeout=None)
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._submit") as mck:
            tp.submit(scenario, False, False, None)
            mck.assert_called_once_with(scenario, force=False, wait=False, timeout=None)
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._submit") as mck:
            tp.submit(sequence, False, False, None)
            mck.assert_called_once_with(sequence, force=False, wait=False, timeout=None)
        with mock.patch("src.taipy.core.task._task_manager._TaskManager._submit") as mck:
            tp.submit(task, False, False, None)
            mck.assert_called_once_with(task, force=False, wait=False, timeout=None)
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._submit") as mck:
            tp.submit(scenario, True, True, 60)
            mck.assert_called_once_with(scenario, force=True, wait=True, timeout=60)
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._submit") as mck:
            tp.submit(sequence, True, True, 60)
            mck.assert_called_once_with(sequence, force=True, wait=True, timeout=60)
        with mock.patch("src.taipy.core.task._task_manager._TaskManager._submit") as mck:
            tp.submit(task, True, True, 60)
            mck.assert_called_once_with(task, force=True, wait=True, timeout=60)

    def test_warning_no_core_service_running(self, scenario):
        _OrchestratorFactory._remove_dispatcher()

        with pytest.warns(ResourceWarning) as warning:
            with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._submit"):
                tp.submit(scenario)

        assert len(warning) == 1
        assert warning[0].message.args[0] == "The Core service is NOT running"

    def test_get_tasks(self):
        with mock.patch("src.taipy.core.task._task_manager._TaskManager._get_all") as mck:
            tp.get_tasks()
            mck.assert_called_once_with()

    def test_get_task(self, task):
        with mock.patch("src.taipy.core.task._task_manager._TaskManager._get") as mck:
            task_id = TaskId("TASK_id")
            tp.get(task_id)
            mck.assert_called_once_with(task_id)

    def test_task_exists(self):
        with mock.patch("src.taipy.core.task._task_manager._TaskManager._exists") as mck:
            task_id = TaskId("TASK_id")
            tp.exists(task_id)
            mck.assert_called_once_with(task_id)

    def test_is_deletable(self, task):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_deletable") as mck:
            scenario_id = ScenarioId("SCENARIO_id")
            tp.is_deletable(scenario_id)
            mck.assert_called_once_with(scenario_id)

        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_deletable") as mck:
            scenario = Scenario("config_id", [], {})
            tp.is_deletable(scenario)
            mck.assert_called_once_with(scenario)

        with mock.patch("src.taipy.core.job._job_manager._JobManager._is_deletable") as mck:
            job_id = JobId("JOB_job_id")
            tp.is_deletable(job_id)
            mck.assert_called_once_with(job_id)

        with mock.patch("src.taipy.core.job._job_manager._JobManager._is_deletable") as mck:
            job = Job("job_id", task, "submit_id", task.id)
            tp.is_deletable(job)
            mck.assert_called_once_with(job)

    def test_is_promotable(self):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_promotable_to_primary") as mck:
            scenario_id = ScenarioId("SCENARIO_id")
            tp.is_promotable(scenario_id)
            mck.assert_called_once_with(scenario_id)

        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._is_promotable_to_primary") as mck:
            scenario = Scenario("config_id", [], {})
            tp.is_promotable(scenario)
            mck.assert_called_once_with(scenario)

    def test_delete_scenario(self):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._hard_delete") as mck:
            scenario_id = ScenarioId("SCENARIO_id")
            tp.delete(scenario_id)
            mck.assert_called_once_with(scenario_id)

    def test_delete(self):
        with mock.patch("src.taipy.core.cycle._cycle_manager._CycleManager._hard_delete") as mck:
            cycle_id = CycleId("CYCLE_id")
            tp.delete(cycle_id)
            mck.assert_called_once_with(cycle_id)

    def test_get_scenarios(self, cycle):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._get_all") as mck:
            tp.get_scenarios()
            mck.assert_called_once_with()
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._get_all_by_cycle") as mck:
            tp.get_scenarios(cycle)
            mck.assert_called_once_with(cycle)
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._get_all_by_tag") as mck:
            tp.get_scenarios(tag="tag")
            mck.assert_called_once_with("tag")

    def test_get_scenario(self, scenario):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._get") as mck:
            scenario_id = ScenarioId("SCENARIO_id")
            tp.get(scenario_id)
            mck.assert_called_once_with(scenario_id)

    def test_scenario_exists(self):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._exists") as mck:
            scenario_id = ScenarioId("SCENARIO_id")
            tp.exists(scenario_id)
            mck.assert_called_once_with(scenario_id)

    def test_get_primary(self, cycle):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._get_primary") as mck:
            tp.get_primary(cycle)
            mck.assert_called_once_with(cycle)

    def test_get_primary_scenarios(self):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._get_primary_scenarios") as mck:
            tp.get_primary_scenarios()
            mck.assert_called_once_with()

    def test_set_primary(self, scenario):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._set_primary") as mck:
            tp.set_primary(scenario)
            mck.assert_called_once_with(scenario)

    def test_tag_and_untag(self, scenario):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._tag") as mck:
            tp.tag(scenario, "tag")
            mck.assert_called_once_with(scenario, "tag")
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._untag") as mck:
            tp.untag(scenario, "tag")
            mck.assert_called_once_with(scenario, "tag")

    def test_compare_scenarios(self, scenario):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._compare") as mck:
            tp.compare_scenarios(scenario, scenario, data_node_config_id="dn")
            mck.assert_called_once_with(scenario, scenario, data_node_config_id="dn")

    def test_subscribe_scenario(self, scenario):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._subscribe") as mck:
            tp.subscribe_scenario(print)
            mck.assert_called_once_with(print, [], None)
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._subscribe") as mck:
            tp.subscribe_scenario(print, scenario=scenario)
            mck.assert_called_once_with(print, [], scenario)

    def test_unsubscribe_scenario(self, scenario):
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._unsubscribe") as mck:
            tp.unsubscribe_scenario(print)
            mck.assert_called_once_with(print, None, None)
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._unsubscribe") as mck:
            tp.unsubscribe_scenario(print, scenario=scenario)
            mck.assert_called_once_with(print, None, scenario)

    def test_subscribe_sequence(self, sequence):
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._subscribe") as mck:
            tp.subscribe_sequence(print)
            mck.assert_called_once_with(print, None, None)
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._subscribe") as mck:
            tp.subscribe_sequence(print, sequence=sequence)
            mck.assert_called_once_with(print, None, sequence)

    def test_unsubscribe_sequence(self, sequence):
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._unsubscribe") as mck:
            tp.unsubscribe_sequence(callback=print)
            mck.assert_called_once_with(print, None, None)
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._unsubscribe") as mck:
            tp.unsubscribe_sequence(callback=print, sequence=sequence)
            mck.assert_called_once_with(print, None, sequence)

    def test_delete_sequence(self):
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._hard_delete") as mck:
            sequence_id = SequenceId("SEQUENCE_id")
            tp.delete(sequence_id)
            mck.assert_called_once_with(sequence_id)

    def test_get_sequence(self, sequence):
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._get") as mck:
            sequence_id = SequenceId("SEQUENCE_id")
            tp.get(sequence_id)
            mck.assert_called_once_with(sequence_id)

    def test_get_sequences(self):
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._get_all") as mck:
            tp.get_sequences()
            mck.assert_called_once_with()

    def test_sequence_exists(self):
        with mock.patch("src.taipy.core.sequence._sequence_manager._SequenceManager._exists") as mck:
            sequence_id = SequenceId("SEQUENCE_id")
            tp.exists(sequence_id)
            mck.assert_called_once_with(sequence_id)

    def test_get_job(self):
        with mock.patch("src.taipy.core.job._job_manager._JobManager._get") as mck:
            job_id = JobId("JOB_id")
            tp.get(job_id)
            mck.assert_called_once_with(job_id)

    def test_get_jobs(self):
        with mock.patch("src.taipy.core.job._job_manager._JobManager._get_all") as mck:
            tp.get_jobs()
            mck.assert_called_once_with()

    def test_job_exists(self):
        with mock.patch("src.taipy.core.job._job_manager._JobManager._exists") as mck:
            job_id = JobId("JOB_id")
            tp.exists(job_id)
            mck.assert_called_once_with(job_id)

    def test_delete_job(self, task):
        with mock.patch("src.taipy.core.job._job_manager._JobManager._delete") as mck:
            job = Job(JobId("job_id"), task, "submit_id", "scenario_id")
            tp.delete_job(job)
            mck.assert_called_once_with(job, False)
        with mock.patch("src.taipy.core.job._job_manager._JobManager._delete") as mck:
            job = Job(JobId("job_id"), task, "submit_id", "scenario_id")
            tp.delete_job(job, False)
            mck.assert_called_once_with(job, False)
        with mock.patch("src.taipy.core.job._job_manager._JobManager._delete") as mck:
            job = Job(JobId("job_id"), task, "submit_id", "scenario_id")
            tp.delete_job(job, True)
            mck.assert_called_once_with(job, True)

    def test_delete_jobs(self):
        with mock.patch("src.taipy.core.job._job_manager._JobManager._delete_all") as mck:
            tp.delete_jobs()
            mck.assert_called_once_with()

    def test_get_latest_job(self, task):
        with mock.patch("src.taipy.core.job._job_manager._JobManager._get_latest") as mck:
            tp.get_latest_job(task)
            mck.assert_called_once_with(task)

    def test_cancel_job(self):
        with mock.patch("src.taipy.core.job._job_manager._JobManager._cancel") as mck:
            tp.cancel_job("job_id")
            mck.assert_called_once_with("job_id")

    def test_block_config_when_core_is_running_in_development_mode(self):
        Config.configure_job_executions(mode=JobConfig._DEVELOPMENT_MODE)

        input_cfg_1 = Config.configure_data_node(id="i1", storage_type="pickle", default_data=1, scope=Scope.SCENARIO)
        output_cfg_1 = Config.configure_data_node(id="o1", storage_type="pickle", scope=Scope.SCENARIO)
        task_cfg_1 = Config.configure_task("t1", print, input_cfg_1, output_cfg_1)
        scenario_cfg_1 = Config.configure_scenario("s1", [task_cfg_1], [], Frequency.DAILY)

        Core().run()

        scenario_1 = tp.create_scenario(scenario_cfg_1)
        tp.submit(scenario_1)

        with pytest.raises(ConfigurationUpdateBlocked):
            Config.configure_scenario("block_scenario", set([task_cfg_1]))

    def test_block_config_when_core_is_running_in_standalone_mode(self):
        Config.configure_job_executions(mode=JobConfig._STANDALONE_MODE)

        input_cfg_1 = Config.configure_data_node(id="i1", storage_type="pickle", default_data=1, scope=Scope.SCENARIO)
        output_cfg_1 = Config.configure_data_node(id="o1", storage_type="pickle", scope=Scope.SCENARIO)
        task_cfg_1 = Config.configure_task("t1", print, input_cfg_1, output_cfg_1)
        scenario_cfg_1 = Config.configure_scenario("s1", [task_cfg_1], [], Frequency.DAILY)

        Core().run()

        scenario_1 = tp.create_scenario(scenario_cfg_1)
        tp.submit(scenario_1, wait=True)

        with pytest.raises(ConfigurationUpdateBlocked):
            Config.configure_scenario("block_scenario", set([task_cfg_1]))

    def test_get_data_node(self, data_node):
        with mock.patch("src.taipy.core.data._data_manager._DataManager._get") as mck:
            tp.get(data_node.id)
            mck.assert_called_once_with(data_node.id)

    def test_get_data_nodes(self):
        with mock.patch("src.taipy.core.data._data_manager._DataManager._get_all") as mck:
            tp.get_data_nodes()
            mck.assert_called_once_with()

    def test_data_node_exists(self):
        with mock.patch("src.taipy.core.data._data_manager._DataManager._exists") as mck:
            data_node_id = DataNodeId("DATANODE_id")
            tp.exists(data_node_id)
            mck.assert_called_once_with(data_node_id)

    def test_get_cycles(self):
        with mock.patch("src.taipy.core.cycle._cycle_manager._CycleManager._get_all") as mck:
            tp.get_cycles()
            mck.assert_called_once_with()

    def test_cycle_exists(self):
        with mock.patch("src.taipy.core.cycle._cycle_manager._CycleManager._exists") as mck:
            cycle_id = CycleId("CYCLE_id")
            tp.exists(cycle_id)
            mck.assert_called_once_with(cycle_id)

    def test_create_global_data_node(self):
        dn_cfg = DataNodeConfig("id", "pickle", Scope.GLOBAL)
        with mock.patch("src.taipy.core.data._data_manager._DataManager._create_and_set") as mck:
            dn = tp.create_global_data_node(dn_cfg)
            mck.assert_called_once_with(dn_cfg, None, None)

        dn = tp.create_global_data_node(dn_cfg)
        assert dn.scope == Scope.GLOBAL
        assert dn.config_id == dn_cfg.id

        # Create a global data node from the same configuration should return the same data node
        dn_2 = tp.create_global_data_node(dn_cfg)
        assert dn_2.id == dn.id

        dn_cfg.scope = Scope.SCENARIO
        with pytest.raises(DataNodeConfigIsNotGlobal):
            tp.create_global_data_node(dn_cfg)

    def test_create_scenario(self, scenario):
        scenario_config = ScenarioConfig("scenario_config")
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._create") as mck:
            tp.create_scenario(scenario_config)
            mck.assert_called_once_with(scenario_config, None, None)
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._create") as mck:
            tp.create_scenario(scenario_config, datetime.datetime(2022, 2, 5))
            mck.assert_called_once_with(scenario_config, datetime.datetime(2022, 2, 5), None)
        with mock.patch("src.taipy.core.scenario._scenario_manager._ScenarioManager._create") as mck:
            tp.create_scenario(scenario_config, datetime.datetime(2022, 2, 5), "displayable_name")
            mck.assert_called_once_with(scenario_config, datetime.datetime(2022, 2, 5), "displayable_name")

    def test_export_scenario_filesystem(self):
        shutil.rmtree("./tmp", ignore_errors=True)

        input_cfg_1 = Config.configure_data_node(id="i1", storage_type="pickle", default_data=1, scope=Scope.SCENARIO)
        output_cfg_1 = Config.configure_data_node(id="o1", storage_type="pickle", scope=Scope.SCENARIO)
        task_cfg_1 = Config.configure_task("t1", print, input_cfg_1, output_cfg_1)
        scenario_cfg_1 = Config.configure_scenario("s1", [task_cfg_1], [], Frequency.DAILY)

        input_cfg_2 = Config.configure_data_node(id="i2", storage_type="pickle", default_data=2, scope=Scope.SCENARIO)
        output_cfg_2 = Config.configure_data_node(id="o2", storage_type="pickle", scope=Scope.SCENARIO)
        task_cfg_2 = Config.configure_task("t2", print, input_cfg_2, output_cfg_2)
        scenario_cfg_2 = Config.configure_scenario("s2", [task_cfg_2], [], Frequency.DAILY)

        scenario_1 = tp.create_scenario(scenario_cfg_1)
        job_1 = tp.submit(scenario_1)[0]

        # Export scenario 1
        tp.export_scenario(scenario_1.id, "./tmp/exp_scenario_1")
        assert sorted(os.listdir("./tmp/exp_scenario_1/data_nodes")) == sorted(
            [f"{scenario_1.i1.id}.json", f"{scenario_1.o1.id}.json"]
        )
        assert sorted(os.listdir("./tmp/exp_scenario_1/tasks")) == sorted([f"{scenario_1.t1.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_1/scenarios")) == sorted([f"{scenario_1.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_1/jobs")) == sorted([f"{job_1.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_1/cycles")) == sorted([f"{scenario_1.cycle.id}.json"])

        scenario_2 = tp.create_scenario(scenario_cfg_2)
        job_2 = tp.submit(scenario_2)[0]

        # Export scenario 2
        scenario_2.export(pathlib.Path.cwd() / "./tmp/exp_scenario_2")
        assert sorted(os.listdir("./tmp/exp_scenario_2/data_nodes")) == sorted(
            [f"{scenario_2.i2.id}.json", f"{scenario_2.o2.id}.json"]
        )
        assert sorted(os.listdir("./tmp/exp_scenario_2/tasks")) == sorted([f"{scenario_2.t2.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_2/scenarios")) == sorted([f"{scenario_2.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_2/jobs")) == sorted([f"{job_2.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_2/cycles")) == sorted([f"{scenario_2.cycle.id}.json"])

        # Export scenario 2 into the folder containing scenario 1 files
        tp.export_scenario(scenario_2.id, "./tmp/exp_scenario_1")
        # Should have the files as scenario 1 only
        assert sorted(os.listdir("./tmp/exp_scenario_1/tasks")) == sorted([f"{scenario_2.t2.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_1/scenarios")) == sorted([f"{scenario_2.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_1/jobs")) == sorted([f"{job_2.id}.json"])
        assert sorted(os.listdir("./tmp/exp_scenario_1/cycles")) == sorted([f"{scenario_2.cycle.id}.json"])

        with pytest.raises(InvalidExportPath):
            tp.export_scenario(scenario_2.id, Config.core.storage_folder)

        shutil.rmtree("./tmp", ignore_errors=True)

    def test_get_parents(self):
        def assert_result_parents_and_expected_parents(parents, expected_parents):
            for key, items in expected_parents.items():
                assert len(parents[key]) == len(expected_parents[key])
                parent_ids = [parent.id for parent in parents[key]]
                assert all([item.id in parent_ids for item in items])

        dn_config_1 = Config.configure_data_node(id="d1", storage_type="in_memory", scope=Scope.SCENARIO)
        dn_config_2 = Config.configure_data_node(id="d2", storage_type="in_memory", scope=Scope.SCENARIO)
        dn_config_3 = Config.configure_data_node(id="d3", storage_type="in_memory", scope=Scope.SCENARIO)
        dn_config_4 = Config.configure_data_node(id="d4", storage_type="in_memory", scope=Scope.SCENARIO)
        task_config_1 = Config.configure_task("t1", print, dn_config_1, dn_config_2)
        task_config_2 = Config.configure_task("t2", print, dn_config_2, dn_config_3)
        scenario_cfg_1 = Config.configure_scenario("s1", [task_config_1, task_config_2], [dn_config_4], Frequency.DAILY)

        scenario = tp.create_scenario(scenario_cfg_1)
        tasks = scenario.tasks

        expected_parents = {
            "scenario": {scenario},
            "task": {tasks["t1"]},
        }
        parents = tp.get_parents(scenario.tasks["t1"].data_nodes["d1"])
        assert_result_parents_and_expected_parents(parents, expected_parents)

        expected_parents = {
            "scenario": {scenario},
            "task": {tasks["t1"], tasks["t2"]},
        }
        parents = tp.get_parents(scenario.tasks["t1"].data_nodes["d2"])
        assert_result_parents_and_expected_parents(parents, expected_parents)

        expected_parents = {"scenario": {scenario}, "task": {tasks["t2"]}}
        parents = tp.get_parents(scenario.tasks["t2"].data_nodes["d3"])
        assert_result_parents_and_expected_parents(parents, expected_parents)

        expected_parents = {"scenario": {scenario}}
        parents = tp.get_parents(scenario.tasks["t1"])
        assert_result_parents_and_expected_parents(parents, expected_parents)

        expected_parents = {"scenario": {scenario}}
        parents = tp.get_parents(scenario.tasks["t2"])
        assert_result_parents_and_expected_parents(parents, expected_parents)

        expected_parents = {"scenario": {scenario}}
        parents = tp.get_parents(scenario.additional_data_nodes["d4"])
        assert_result_parents_and_expected_parents(parents, expected_parents)

        expected_parents = {}
        parents = tp.get_parents(scenario)
        assert_result_parents_and_expected_parents(parents, expected_parents)

        expected_parents = {}
        parents = tp.get_parents(scenario.cycle)
        assert_result_parents_and_expected_parents(parents, expected_parents)

    def test_get_cycles_scenarios(self):
        scenario_cfg_1 = Config.configure_scenario(
            "s1",
            set(),
            set(),
            Frequency.DAILY,
        )
        scenario_cfg_2 = Config.configure_scenario("s2", set(), set(), Frequency.WEEKLY)
        scenario_cfg_3 = Config.configure_scenario("s3", set(), set(), Frequency.MONTHLY)
        scenario_cfg_4 = Config.configure_scenario("s4", set(), set(), Frequency.YEARLY)
        scenario_cfg_5 = Config.configure_scenario("s5", set(), set(), None)

        now = datetime.datetime.now()
        scenario_1_1 = tp.create_scenario(scenario_cfg_1, now)
        scenario_1_2 = tp.create_scenario(scenario_cfg_1, datetime.datetime.now())
        scenario_1_3 = tp.create_scenario(scenario_cfg_1, now + datetime.timedelta(days=1))
        scenario_1_4 = tp.create_scenario(scenario_cfg_1, now + datetime.timedelta(days=8))
        scenario_1_5 = tp.create_scenario(scenario_cfg_1, now + datetime.timedelta(days=25))
        scenario_2 = tp.create_scenario(scenario_cfg_2)
        scenario_3 = tp.create_scenario(scenario_cfg_3)
        scenario_4 = tp.create_scenario(scenario_cfg_4)
        scenario_5_1 = tp.create_scenario(scenario_cfg_5)
        scenario_5_2 = tp.create_scenario(scenario_cfg_5)
        scenario_5_3 = tp.create_scenario(scenario_cfg_5)

        expected_cycles_scenarios = {
            scenario_1_1.cycle: [scenario_1_1.id, scenario_1_2.id],
            scenario_1_3.cycle: [scenario_1_3.id],
            scenario_1_4.cycle: [scenario_1_4.id],
            scenario_1_5.cycle: [scenario_1_5.id],
            scenario_2.cycle: [scenario_2.id],
            scenario_3.cycle: [scenario_3.id],
            scenario_4.cycle: [scenario_4.id],
            None: [scenario_5_1.id, scenario_5_2.id, scenario_5_3.id],
        }

        cycles_scenarios = tp.get_cycles_scenarios()

        assert expected_cycles_scenarios.keys() == cycles_scenarios.keys()
        for cycle, scenarios in cycles_scenarios.items():
            expected_scenarios = expected_cycles_scenarios[cycle]
            assert sorted([scenario.id for scenario in scenarios]) == sorted(expected_scenarios)

    def test_get_entities_by_config_id(self):
        scenario_config_1 = Config.configure_scenario("s1", sequence_configs=[])
        scenario_config_2 = Config.configure_scenario("s2", sequence_configs=[])

        s_1_1 = tp.create_scenario(scenario_config_1)
        s_1_2 = tp.create_scenario(scenario_config_1)
        s_1_3 = tp.create_scenario(scenario_config_1)

        assert len(tp.get_scenarios()) == 3

        s_2_1 = tp.create_scenario(scenario_config_2)
        s_2_2 = tp.create_scenario(scenario_config_2)
        assert len(tp.get_scenarios()) == 5

        s1_scenarios = tp.get_entities_by_config_id(scenario_config_1.id)
        assert len(s1_scenarios) == 3
        assert sorted([s_1_1.id, s_1_2.id, s_1_3.id]) == sorted([scenario.id for scenario in s1_scenarios])

        s2_scenarios = tp.get_entities_by_config_id(scenario_config_2.id)
        assert len(s2_scenarios) == 2
        assert sorted([s_2_1.id, s_2_2.id]) == sorted([scenario.id for scenario in s2_scenarios])

    def test_get_entities_by_config_id_in_multiple_versions_environment(self):
        scenario_config_1 = Config.configure_scenario("s1", sequence_configs=[])
        scenario_config_2 = Config.configure_scenario("s2", sequence_configs=[])

        _VersionManager._set_experiment_version("1.0")
        tp.create_scenario(scenario_config_1)
        tp.create_scenario(scenario_config_1)
        tp.create_scenario(scenario_config_1)
        tp.create_scenario(scenario_config_2)
        tp.create_scenario(scenario_config_2)
        assert len(tp.get_scenarios()) == 5
        assert len(tp.get_entities_by_config_id(scenario_config_1.id)) == 3
        assert len(tp.get_entities_by_config_id(scenario_config_2.id)) == 2

        _VersionManager._set_experiment_version("2.0")
        tp.create_scenario(scenario_config_1)
        tp.create_scenario(scenario_config_1)
        tp.create_scenario(scenario_config_1)
        tp.create_scenario(scenario_config_2)
        tp.create_scenario(scenario_config_2)
        assert len(tp.get_scenarios()) == 5
        assert len(tp.get_entities_by_config_id(scenario_config_1.id)) == 3
        assert len(tp.get_entities_by_config_id(scenario_config_2.id)) == 2



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import json
from typing import Dict, List, Optional, Tuple

from taipy.logger._taipy_logger import _TaipyLogger

__logger = _TaipyLogger._get_logger()


def __update_parent_ids(entity: Dict, data: Dict) -> Dict:
    # parent_ids was not present in 2.0, need to be search for in tasks
    parent_ids = entity.get("parent_ids", [])
    if not parent_ids:
        parent_ids = __search_parent_ids(entity["id"], data)
    entity["parent_ids"] = parent_ids

    return entity


def __update_config_parent_ids(id: str, entity: Dict, entity_type: str, config: Dict) -> Dict:
    # parent_ids was not present in 2.0, need to be search for in tasks
    parent_ids = entity.get("parent_ids", [])

    if not parent_ids:
        parent_ids = __search_parent_config(id, config, entity_type)
    entity["parent_ids"] = parent_ids

    return entity


def __search_parent_ids(entity_id: str, data: Dict) -> List:
    parents = []
    entity_type = entity_id.split("_", 1)[0]

    for _id, entity_data in data.items():
        entity_data = entity_data["data"]
        if entity_type == "DATANODE" and "TASK" in _id:
            if entity_id in entity_data["input_ids"] or entity_id in entity_data["output_ids"]:
                parents.append(_id)

        if entity_type == "TASK" and "SCENARIO" in _id:
            if entity_id in entity_data["tasks"]:
                parents.append(_id)
    return parents


def __search_parent_config(entity_id: str, config: Dict, entity_type: str) -> List:
    parents = []
    possible_parents = "TASK" if entity_type == "DATA_NODE" else "SCENARIO"
    data = config[possible_parents]

    for _id, entity_data in data.items():
        section_id = f"{entity_id}:SECTION"
        if entity_type == "DATANODE" and possible_parents == "TASK":
            if section_id in entity_data["input_ids"] or section_id in entity_data["output_ids"]:
                parents.append(section_id)

        if entity_type == "TASK" and possible_parents == "SCENARIO":
            if section_id in entity_data["tasks"]:
                parents.append(section_id)
    return parents


def __fetch_tasks_from_pipelines(pipelines: List, data: Dict) -> List:
    tasks = []
    for pipeline in pipelines:
        pipeline_data = data[pipeline]["data"]
        tasks.extend(pipeline_data["tasks"])
    return tasks


def __migrate_subscriber(fct_module, fct_name):
    """Rename scheduler by orchestrator on old jobs. Used to migrate from <=2.2 to >=2.3 version."""
    if fct_module == "taipy.core._scheduler._scheduler":
        fct_module = fct_module.replace("_scheduler", "_orchestrator")
        fct_name = fct_name.replace("_Scheduler", "_Orchestrator")
    return fct_module, fct_name


def __migrate_scenario(scenario: Dict, data: Dict) -> Dict:
    # pipelines were replaced by tasks
    scenario["tasks"] = __fetch_tasks_from_pipelines(scenario["pipelines"], data)

    # pipeline attribute not removed in 3.0
    scenario["pipelines"] = None

    # additional_data_nodes attribute added
    scenario["additional_data_nodes"] = []

    return scenario


def __is_cacheable(task: Dict, data: Dict) -> bool:
    output_ids = task.get("output_ids", []) or task.get("outputs", [])  # output_ids is on entity, outputs is on config

    for output_id in output_ids:
        if output_id.endswith(":SECTION"):  # Get the config_id if the task is a Config
            output_id = output_id.split(":")[0]
        dn = data.get(output_id, {})
        if "data" in dn:
            dn = dn.get("data", {})

        if "cacheable" not in dn or not dn["cacheable"] or dn["cacheable"] == "False:bool":
            return False
    return True


def __migrate_task(task: Dict, data: Dict, is_entity: bool = True) -> Dict:
    if is_entity:
        # parent_id has been renamed to owner_id
        try:
            task["owner_id"] = task["parent_id"]
            del task["parent_id"]
        except KeyError:
            pass

        # properties was not present in 2.0
        task["properties"] = task.get("properties", {})

    # skippable was not present in 2.0
    task["skippable"] = task.get("skippable", False) or __is_cacheable(task, data)

    return task


def __migrate_task_entity(task: Dict, data: Dict) -> Dict:
    task = __update_parent_ids(task, data)
    return __migrate_task(task, data)


def __migrate_task_config(task: Dict, config: Dict) -> Dict:
    task = __migrate_task(task, config["DATA_NODE"], False)

    # Convert the skippable boolean to a string if needed
    if isinstance(task.get("skippable"), bool):
        task["skippable"] = str(task["skippable"]) + ":bool"
    return task


def __update_scope(scope: str):
    if scope in "<Scope.SCENARIO: 2>":
        return "<Scope.SCENARIO: 1>"
    elif scope == "<Scope.CYCLE: 3>":
        return "<Scope.CYCLE: 2>"
    elif scope == "<Scope.GLOBAL: 4>":
        return "<Scope.GLOBAL: 3>"
    return scope


def __migrate_datanode(datanode: Dict) -> Dict:
    # cacheable was removed in after 2.0
    _ = datanode.pop("cacheable", False)

    # job_ids was replaced by edits
    if "job_ids" in datanode:
        datanode["edits"] = [{"job_id": job, "timestamp": datanode["last_edit_date"]} for job in datanode["job_ids"]]
    elif "edits" in datanode:
        # make sure timestamp inside edits is a string
        edits = []
        for edit in datanode["edits"]:
            timestamp = edit.get("timestamp")
            if isinstance(timestamp, dict):
                timestamp = timestamp.get("__value__")
            new_edit = {"timestamp": timestamp}
            if "job_id" in edit:
                new_edit["job_id"] = edit["job_id"]
            edits.append(new_edit)
        datanode["edits"] = edits

    # parent_id has been renamed to owner_id
    try:
        datanode["owner_id"] = datanode["parent_id"]
        del datanode["parent_id"]
    except KeyError:
        pass

    # Update Scope enum after Pipeline removal
    datanode["scope"] = __update_scope(datanode["scope"])

    if "last_edit_date" not in datanode:
        datanode["last_edit_date"] = datanode.get("last_edition_date")
        if "last_edition_date" in datanode:
            del datanode["last_edition_date"]

    if "edit_in_progress" not in datanode:
        datanode["edit_in_progress"] = datanode.get("edition_in_progress")
        if "edition_in_progress" in datanode:
            del datanode["edition_in_progress"]

    return datanode


def __migrate_datanode_entity(datanode: Dict, data: Dict) -> Dict:
    datanode = __update_parent_ids(datanode, data)
    return __migrate_datanode(datanode)


def __migrate_datanode_config(datanode: Dict) -> Dict:
    if datanode["storage_type"] in ["csv", "json"]:
        datanode["encoding"] = "utf-8"
    return datanode


def __migrate_job(job: Dict) -> Dict:
    # submit_entity_id was not present before 3.0
    job["submit_entity_id"] = job.get("submit_entity_id", None)
    if "subscribers" in job:
        for sub in job["subscribers"]:
            sub["fct_module"], sub["fct_name"] = __migrate_subscriber(sub["fct_module"], sub["fct_name"])
    return job


def __migrate_global_config(config: Dict):
    fields_to_remove = ["clean_entities_enabled"]
    fields_to_move = ["root_folder", "storage_folder", "repository_type", "read_entity_retry"]

    for field in fields_to_remove:
        if field in config["TAIPY"]:
            del config["TAIPY"][field]
    try:
        for field in fields_to_move:
            if field not in config["CORE"]:
                config["CORE"][field] = config["TAIPY"][field]
                del config["TAIPY"][field]
    except KeyError:
        pass

    return config


def __migrate_version(version: Dict) -> Dict:
    config_str = version["config"]

    # Remove PIPELINE scope
    config_str = config_str.replace("PIPELINE:SCOPE", "SCENARIO:SCOPE")
    config = json.loads(config_str)

    # remove unused fields and move others from TAIPY to CORE section
    config = __migrate_global_config(config)

    # replace pipelines for tasks
    pipelines_section = config["PIPELINE"]
    for id, content in config["SCENARIO"].items():
        tasks = []
        for _pipeline in content["pipelines"]:
            pipeline_id = _pipeline.split(":")[0]
            tasks = pipelines_section[pipeline_id]["tasks"]
        config["SCENARIO"][id]["tasks"] = tasks
        del config["SCENARIO"][id]["pipelines"]

    for id, content in config["TASK"].items():
        config["TASK"][id] = __migrate_task_config(content, config)

    for id, content in config["DATA_NODE"].items():
        config["DATA_NODE"][id] = __migrate_datanode_config(content)

    del config["PIPELINE"]

    version["config"] = json.dumps(config, ensure_ascii=False, indent=0)
    return version


def __migrate_entities(entity_type: str, data: Dict) -> Dict:
    migration_fct = FCT_MIGRATION_MAP.get(entity_type)
    _entities = {k: data[k] for k in data if entity_type in k}

    for k, v in _entities.items():
        if entity_type in ["JOB", "VERSION"]:
            v["data"] = migration_fct(v["data"])  # type: ignore
        else:
            v["data"] = migration_fct(v["data"], data)  # type: ignore
        data[k] = v
    return data


FCT_MIGRATION_MAP = {
    "SCENARIO": __migrate_scenario,
    "TASK": __migrate_task_entity,
    "DATANODE": __migrate_datanode_entity,
    "JOB": __migrate_job,
    "VERSION": __migrate_version,
}


def _migrate(entities: Dict, versions: Optional[Dict] = None) -> Tuple[Dict, Dict]:
    __logger.info("Migrating SCENARIOS")
    entities = __migrate_entities("SCENARIO", entities)

    __logger.info("Migrating TASKS")
    entities = __migrate_entities("TASK", entities)

    __logger.info("Migrating DATANODES")
    entities = __migrate_entities("DATANODE", entities)

    __logger.info("Migrating JOBS")
    entities = __migrate_entities("JOB", entities)

    __logger.info("Migrating VERSION")
    if versions:
        versions = __migrate_entities("VERSION", versions)
    else:
        entities = __migrate_entities("VERSION", entities)
        versions = {}
    return entities, versions



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .._repository._sql_repository import _SQLRepository
from ..exceptions.exceptions import ModelNotFound, VersionIsNotProductionVersion
from ._version_converter import _VersionConverter
from ._version_model import _VersionModel
from ._version_repository_interface import _VersionRepositoryInterface


class _VersionSQLRepository(_SQLRepository, _VersionRepositoryInterface):
    def __init__(self):
        super().__init__(model_type=_VersionModel, converter=_VersionConverter)

    def _set_latest_version(self, version_number):
        if old_latest := self.db.query(self.model_type).filter_by(is_latest=True).first():
            old_latest.is_latest = False

        version = self.__get_by_id(version_number)
        version.is_latest = True

        self.db.commit()

    def _get_latest_version(self):
        if latest := self.db.query(self.model_type).filter_by(is_latest=True).first():
            return latest.id
        return ""

    def _set_development_version(self, version_number):
        if old_development := self.db.query(self.model_type).filter_by(is_development=True).first():
            old_development.is_development = False

        version = self.__get_by_id(version_number)
        version.is_development = True

        self._set_latest_version(version_number)

        self.db.commit()

    def _get_development_version(self):
        if development := self.db.query(self.model_type).filter_by(is_development=True).first():
            return development.id
        raise ModelNotFound(self.model_type, "")

    def _set_production_version(self, version_number):
        version = self.__get_by_id(version_number)
        version.is_production = True

        self._set_latest_version(version_number)

        self.db.commit()

    def _get_production_versions(self):
        if productions := self.db.query(self.model_type).filter_by(is_production=True).all():
            return [p.id for p in productions]
        return []

    def _delete_production_version(self, version_number):
        version = self.__get_by_id(version_number)

        if not version or not version.is_production:
            raise VersionIsNotProductionVersion(f"Version '{version_number}' is not a production version.")
        version.is_production = False

        self.db.commit()

    def __get_by_id(self, version_id):
        return self.db.query(self.model_type).filter_by(id=version_id).first()



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""
Package for notifications about changes on `Core^` service entities.


The Core service generates `Event^` objects to track changes on entities.
These events are then relayed to a `Notifier^`, which handles the dispatch
to consumers interested in specific event topics.

To subscribe, a consumer needs to invoke the `Notifier.register()^` method.
This call will yield a `RegistrationId^` and a dedicated event queue for
receiving notifications.

To handle notifications, an event consumer (e.g., the `CoreEventConsumerBase^`
object) must be instantiated with an associated event queue.
"""

from ._registration import _Registration
from ._topic import _Topic
from .core_event_consumer import CoreEventConsumerBase
from .event import _ENTITY_TO_EVENT_ENTITY_TYPE, Event, EventEntityType, EventOperation
from .notifier import Notifier, _publish_event
from .registration_id import RegistrationId



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from sqlalchemy import JSON, Boolean, Column, String, Table

from .._repository._base_taipy_model import _BaseModel
from .._repository.db._sql_base_model import mapper_registry
from ..cycle.cycle_id import CycleId
from ..data.data_node_id import DataNodeId
from ..task.task_id import TaskId
from .scenario_id import ScenarioId


@mapper_registry.mapped
@dataclass
class _ScenarioModel(_BaseModel):
    __table__ = Table(
        "scenario",
        mapper_registry.metadata,
        Column("id", String, primary_key=True),
        Column("config_id", String),
        Column("tasks", JSON),
        Column("additional_data_nodes", JSON),
        Column("properties", JSON),
        Column("creation_date", String),
        Column("primary_scenario", Boolean),
        Column("subscribers", JSON),
        Column("tags", JSON),
        Column("version", String),
        Column("sequences", JSON),
        Column("cycle", String),
    )
    id: ScenarioId
    config_id: str
    tasks: List[TaskId]
    additional_data_nodes: List[DataNodeId]
    properties: Dict[str, Any]
    creation_date: str
    primary_scenario: bool
    subscribers: List[Dict]
    tags: List[str]
    version: str
    sequences: Optional[Dict[str, Dict]] = None
    cycle: Optional[CycleId] = None

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        return _ScenarioModel(
            id=data["id"],
            config_id=data["config_id"],
            tasks=data.get("tasks", None),
            additional_data_nodes=data.get("additional_data_nodes", None),
            properties=data["properties"],
            creation_date=data["creation_date"],
            primary_scenario=data["primary_scenario"],
            subscribers=data["subscribers"],
            tags=data["tags"],
            version=data["version"],
            sequences=data.get("sequences", None),
            cycle=CycleId(data["cycle"]) if "cycle" in data else None,
        )



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import dataclasses
import enum
from typing import Any, Dict


class _BaseModel:
    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def to_dict(self) -> Dict[str, Any]:
        model_dict = {**dataclasses.asdict(self)}

        for k, v in model_dict.items():
            if isinstance(v, enum.Enum):
                model_dict[k] = repr(v)
        return model_dict



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
from copy import copy
from typing import Any, Dict, Optional, Union

from taipy.config import Config, UniqueSection
from taipy.config._config import _Config
from taipy.config.common._config_blocker import _ConfigBlocker
from taipy.config.common._template_handler import _TemplateHandler as _tpl

from .._init_version import _read_version
from ..exceptions.exceptions import ConfigCoreVersionMismatched


class CoreSection(UniqueSection):
    """
    Configuration parameters for running the `Core^` service.

    Attributes:
        root_folder (str): Path of the base folder for the taipy application. The default value is "./taipy/"
        storage_folder (str): Folder name used to store Taipy data. The default value is ".data/". It is used in
            conjunction with the *root_folder* attribute. That means the storage path is <root_folder><storage_folder>
            (The default path is "./taipy/.data/").
        repository_type (str): Type of the repository to be used to store Taipy data. The default value is
            "filesystem".
        repository_properties (Dict[str, Union[str, int]]): A dictionary of additional properties to be used by the
            repository.
        read_entity_retry (int): Number of retries to read an entity from the repository before return failure.
            The default value is 3.
        mode (str): The Taipy operating mode. By default, the `Core^` service runs in "development" mode.
            An "experiment" and a "production" mode are also available. Please refer to the
            [Versioning management](../core/versioning/index.md) documentation page for more details.
        version_number (str)): The identifier of the user application version. Please refer to the
            [Versioning management](../core/versioning/index.md) documentation page for more details.
        force (bool): If True, force the application run even if there are some conflicts in the
            configuration.
        core_version (str): The Taipy Core package version.
        **properties (dict[str, any]): A dictionary of additional properties.
    """

    name = "CORE"

    _ROOT_FOLDER_KEY = "root_folder"
    _DEFAULT_ROOT_FOLDER = "./taipy/"

    _STORAGE_FOLDER_KEY = "storage_folder"
    _DEFAULT_STORAGE_FOLDER = ".data/"

    _REPOSITORY_TYPE_KEY = "repository_type"
    _DEFAULT_REPOSITORY_TYPE = "filesystem"
    _REPOSITORY_PROPERTIES_KEY = "repository_properties"
    _DEFAULT_REPOSITORY_PROPERTIES: Dict = dict()

    _READ_ENTITY_RETRY_KEY = "read_entity_retry"
    _DEFAULT_READ_ENTITY_RETRY = 1

    _MODE_KEY = "mode"
    _DEVELOPMENT_MODE = "development"
    _EXPERIMENT_MODE = "experiment"
    _PRODUCTION_MODE = "production"
    _DEFAULT_MODE = _DEVELOPMENT_MODE

    _VERSION_NUMBER_KEY = "version_number"
    _DEFAULT_VERSION_NUMBER = ""

    _FORCE_KEY = "force"
    _DEFAULT_FORCE = False

    _CORE_VERSION_KEY = "core_version"
    _CURRENT_CORE_VERSION = _read_version()

    def __init__(
        self,
        root_folder: Optional[str] = None,
        storage_folder: Optional[str] = None,
        repository_type: Optional[str] = None,
        repository_properties: Optional[Dict[str, Union[str, int]]] = None,
        read_entity_retry: Optional[int] = None,
        mode: Optional[str] = None,
        version_number: Optional[str] = None,
        force: Optional[bool] = None,
        core_version: Optional[str] = None,
        **properties,
    ):
        self._root_folder = root_folder
        self._storage_folder = storage_folder
        self._repository_type = repository_type
        self._repository_properties = repository_properties or {}
        self._read_entity_retry = (
            read_entity_retry if read_entity_retry is not None else self._DEFAULT_READ_ENTITY_RETRY
        )
        self.mode = mode or self._DEFAULT_MODE
        self.version_number = version_number or self._DEFAULT_VERSION_NUMBER
        self.force = force or self._DEFAULT_FORCE

        self._check_compatibility(core_version)
        self._core_version = core_version

        super().__init__(**properties)

    def __copy__(self):
        return CoreSection(
            self.root_folder,
            self.storage_folder,
            self.repository_type,
            self.repository_properties,
            self.read_entity_retry,
            self.mode,
            self.version_number,
            self.force,
            self._core_version,
            **copy(self._properties),
        )

    @property
    def storage_folder(self):
        return _tpl._replace_templates(self._storage_folder)

    @storage_folder.setter  # type: ignore
    @_ConfigBlocker._check()
    def storage_folder(self, val):
        self._storage_folder = val

    @property
    def root_folder(self):
        return _tpl._replace_templates(self._root_folder)

    @root_folder.setter  # type: ignore
    @_ConfigBlocker._check()
    def root_folder(self, val):
        self._root_folder = val

    @property
    def repository_type(self):
        return _tpl._replace_templates(self._repository_type)

    @repository_type.setter  # type: ignore
    @_ConfigBlocker._check()
    def repository_type(self, val):
        self._repository_type = val

    @property
    def repository_properties(self):
        return (
            {k: _tpl._replace_templates(v) for k, v in self._repository_properties.items()}
            if self._repository_properties
            else self._DEFAULT_REPOSITORY_PROPERTIES.copy()
        )

    @repository_properties.setter  # type: ignore
    @_ConfigBlocker._check()
    def repository_properties(self, val):
        self._repository_properties = val

    @property
    def read_entity_retry(self):
        return _tpl._replace_templates(self._read_entity_retry)

    @read_entity_retry.setter  # type: ignore
    @_ConfigBlocker._check()
    def read_entity_retry(self, val):
        self._read_entity_retry = val

    @classmethod
    def default_config(cls):
        return CoreSection(
            cls._DEFAULT_ROOT_FOLDER,
            cls._DEFAULT_STORAGE_FOLDER,
            cls._DEFAULT_REPOSITORY_TYPE,
            cls._DEFAULT_REPOSITORY_PROPERTIES,
            cls._DEFAULT_READ_ENTITY_RETRY,
            cls._DEFAULT_MODE,
            cls._DEFAULT_VERSION_NUMBER,
            cls._DEFAULT_FORCE,
            cls._CURRENT_CORE_VERSION,
        )

    def _clean(self):
        self._root_folder = self._DEFAULT_ROOT_FOLDER
        self._storage_folder = self._DEFAULT_STORAGE_FOLDER
        self._repository_type = self._DEFAULT_REPOSITORY_TYPE
        self._repository_properties = self._DEFAULT_REPOSITORY_PROPERTIES.copy()
        self._read_entity_retry = self._DEFAULT_READ_ENTITY_RETRY
        self.mode = self._DEFAULT_MODE
        self.version_number = self._DEFAULT_VERSION_NUMBER
        self.force = self._DEFAULT_FORCE
        self._core_version = self._CURRENT_CORE_VERSION
        self._properties.clear()

    def _to_dict(self):
        as_dict = {}
        if self._root_folder:
            as_dict[self._ROOT_FOLDER_KEY] = self._root_folder
        if self._storage_folder:
            as_dict[self._STORAGE_FOLDER_KEY] = self._storage_folder
        if self._repository_type:
            as_dict[self._REPOSITORY_TYPE_KEY] = self._repository_type
        if self._repository_properties:
            as_dict[self._REPOSITORY_PROPERTIES_KEY] = self._repository_properties
        if self._read_entity_retry is not None:
            as_dict[self._READ_ENTITY_RETRY_KEY] = self._read_entity_retry
        if self.mode is not None:
            as_dict[self._MODE_KEY] = self.mode
        if self.version_number is not None:
            as_dict[self._VERSION_NUMBER_KEY] = self.version_number
        if self.force is not None:
            as_dict[self._FORCE_KEY] = self.force
        if self._core_version is not None:
            as_dict[self._CORE_VERSION_KEY] = self._core_version
        as_dict.update(self._properties)
        return as_dict

    @classmethod
    def _from_dict(cls, as_dict: Dict[str, Any], id=None, config: Optional[_Config] = None):
        root_folder = as_dict.pop(cls._ROOT_FOLDER_KEY, None)
        storage_folder = as_dict.pop(cls._STORAGE_FOLDER_KEY, None)
        repository_type = as_dict.pop(cls._REPOSITORY_TYPE_KEY, None)
        repository_properties = as_dict.pop(cls._REPOSITORY_PROPERTIES_KEY, None)
        read_entity_retry = as_dict.pop(cls._READ_ENTITY_RETRY_KEY, None)
        mode = as_dict.pop(cls._MODE_KEY, None)
        version_nb = as_dict.pop(cls._VERSION_NUMBER_KEY, None)
        force = as_dict.pop(cls._FORCE_KEY, None)
        core_version = as_dict.pop(cls._CORE_VERSION_KEY, None)
        return CoreSection(
            root_folder,
            storage_folder,
            repository_type,
            repository_properties,
            read_entity_retry,
            mode,
            version_nb,
            force,
            core_version,
            **as_dict,
        )

    def _update(self, as_dict: Dict[str, Any]):
        root_folder = _tpl._replace_templates(as_dict.pop(self._ROOT_FOLDER_KEY, self._root_folder))
        if self._root_folder != root_folder:
            self._root_folder = root_folder

        storage_folder = _tpl._replace_templates(as_dict.pop(self._STORAGE_FOLDER_KEY, self._storage_folder))
        if self._storage_folder != storage_folder:
            self._storage_folder = storage_folder

        repository_type = _tpl._replace_templates(as_dict.pop(self._REPOSITORY_TYPE_KEY, self._repository_type))
        if self._repository_type != repository_type:
            self._repository_type = repository_type

        repository_properties = _tpl._replace_templates(
            as_dict.pop(self._REPOSITORY_PROPERTIES_KEY, self._repository_properties)
        )
        self._repository_properties.update(repository_properties)

        read_entity_retry = _tpl._replace_templates(as_dict.pop(self._READ_ENTITY_RETRY_KEY, self._read_entity_retry))
        if self._read_entity_retry != read_entity_retry:
            self._read_entity_retry = read_entity_retry

        mode = _tpl._replace_templates(as_dict.pop(self._MODE_KEY, self.mode))
        if self.mode != mode:
            self.mode = mode

        version_number = _tpl._replace_templates(as_dict.pop(self._VERSION_NUMBER_KEY, self.version_number))
        if self.version_number != version_number:
            self.version_number = version_number

        force = _tpl._replace_templates(as_dict.pop(self._FORCE_KEY, self.force))
        if self.force != force:
            self.force = force

        core_version = as_dict.pop(self._CORE_VERSION_KEY, None)
        self._check_compatibility(core_version)

        self._properties.update(as_dict)

    @classmethod
    def _check_compatibility(cls, core_version):
        if not core_version:
            return

        version_pattern = r"^(\d+)\.(\d+)\.(\d+)$"
        dev_version_pattern = r"^(\d+)\.(\d+)\.(\d+).(\w*)$"

        installed_match = re.match(version_pattern, cls._CURRENT_CORE_VERSION) or re.match(
            dev_version_pattern, cls._CURRENT_CORE_VERSION
        )
        required_match = re.match(version_pattern, core_version) or re.match(dev_version_pattern, core_version)
        if required_match and installed_match:
            installed_group = installed_match.groups()
            required_group = required_match.groups()
            installed_major, installed_minor = installed_group[0], installed_group[1]
            required_major, required_minor = required_group[0], required_group[1]

            if required_major != installed_major or required_minor != installed_minor:
                raise ConfigCoreVersionMismatched(core_version, cls._CURRENT_CORE_VERSION)

    @staticmethod
    def _configure(
        root_folder: Optional[str] = None,
        storage_folder: Optional[str] = None,
        repository_type: Optional[str] = None,
        repository_properties: Optional[Dict[str, Union[str, int]]] = None,
        read_entity_retry: Optional[int] = None,
        mode: Optional[str] = None,
        version_number: Optional[str] = None,
        force: Optional[bool] = None,
        **properties,
    ) -> "CoreSection":
        """Configure the Core service.

        Parameters:
            root_folder (Optional[str]): Path of the base folder for the taipy application.
                The default value is "./taipy/"
            storage_folder (Optional[str]): Folder name used to store Taipy data. The default value is ".data/".
                It is used in conjunction with the `root_folder` field. That means the storage path is
                <root_folder><storage_folder> (The default path is "./taipy/.data/").
            repository_type (Optional[str]): The type of the repository to be used to store Taipy data.
                The default value is "filesystem".
            repository_properties (Optional[Dict[str, Union[str, int]]]): A dictionary of additional properties
                to be used by the repository.
            read_entity_retry (Optional[int]): Number of retries to read an entity from the repository
                before return failure. The default value is 3.
            mode (Optional[str]): Indicates the mode of the version management system.
                Possible values are *"development"*, *"experiment"*, or *"production"*.
            version_number (Optional[str]): The string identifier of the version.
                 In development mode, the version number is ignored.
            force (Optional[bool]): If True, Taipy will override a version even if the configuration
                has changed and run the application.
            **properties (Dict[str, Any]): A keyworded variable length list of additional arguments configure the
                behavior of the `Core^` service.
        Returns:
            The Core configuration.
        """
        section = CoreSection(
            root_folder=root_folder,
            storage_folder=storage_folder,
            repository_type=repository_type,
            repository_properties=repository_properties,
            read_entity_retry=read_entity_retry,
            mode=mode,
            version_number=version_number,
            force=force,
            core_version=_read_version(),
            **properties,
        )
        Config._register(section)
        return Config.unique_sections[CoreSection.name]



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime, timedelta
from inspect import isclass
from typing import Any, Dict, List, Optional, Set

from taipy.config.common.scope import Scope

from .._version._version_manager_factory import _VersionManagerFactory
from ..common._mongo_connector import _connect_mongodb
from ..exceptions.exceptions import InvalidCustomDocument, MissingRequiredProperty
from .data_node import DataNode
from .data_node_id import DataNodeId, Edit


class MongoCollectionDataNode(DataNode):
    """Data Node stored in a Mongo collection.

    Attributes:
        config_id (str): Identifier of the data node configuration. It must be a valid Python
            identifier.
        scope (Scope^): The scope of this data node.
        id (str): The unique identifier of this data node.
        name (str): A user-readable name of this data node.
        owner_id (str): The identifier of the owner (sequence_id, scenario_id, cycle_id) or
            None.
        parent_ids (Optional[Set[str]]): The identifiers of the parent tasks or `None`.
        last_edit_date (datetime): The date and time of the last modification.
        edits (List[Edit^]): The ordered list of edits for that job.
        version (str): The string indicates the application version of the data node to instantiate. If not provided,
            the current version is used.
        validity_period (Optional[timedelta]): The duration implemented as a timedelta since the last edit date for
            which the data node can be considered up-to-date. Once the validity period has passed, the data node is
            considered stale and relevant tasks will run even if they are skippable (see the
            [Task management page](../core/entities/task-mgt.md) for more details).
            If _validity_period_ is set to `None`, the data node is always up-to-date.
        edit_in_progress (bool): True if a task computing the data node has been submitted
            and not completed yet. False otherwise.
        editor_id (Optional[str]): The identifier of the user who is currently editing the data node.
        editor_expiration_date (Optional[datetime]): The expiration date of the editor lock.
        properties (dict[str, Any]): A dictionary of additional properties. Note that the
            _properties_ parameter must at least contain an entry for _"db_name"_ and _"collection_name"_:

            - _"db_name"_ `(str)`: The database name.\n
            - _"collection_name"_ `(str)`: The collection in the database to read from and to write the data to.\n
            - _"custom_document"_ `(Any)`: The custom document class to store, encode, and decode data when reading and
                writing to a Mongo collection.\n
            - _"db_username"_ `(str)`: The database username.\n
            - _"db_password"_ `(str)`: The database password.\n
            - _"db_host"_ `(str)`: The database host. The default value is _"localhost"_.\n
            - _"db_port"_ `(int)`: The database port. The default value is 27017.\n
            - _"db_driver"_ `(str)`: The database driver.\n
            - _"db_extra_args"_ `(Dict[str, Any])`: A dictionary of additional arguments to be passed into database
                connection string.\n
    """

    __STORAGE_TYPE = "mongo_collection"

    __DB_NAME_KEY = "db_name"
    __COLLECTION_KEY = "collection_name"
    __DB_USERNAME_KEY = "db_username"
    __DB_PASSWORD_KEY = "db_password"
    __DB_HOST_KEY = "db_host"
    __DB_PORT_KEY = "db_port"
    __DB_EXTRA_ARGS_KEY = "db_extra_args"
    __DB_DRIVER_KEY = "db_driver"

    __DB_HOST_DEFAULT = "localhost"
    __DB_PORT_DEFAULT = 27017

    _CUSTOM_DOCUMENT_PROPERTY = "custom_document"
    _REQUIRED_PROPERTIES: List[str] = [
        __DB_NAME_KEY,
        __COLLECTION_KEY,
    ]

    def __init__(
        self,
        config_id: str,
        scope: Scope,
        id: Optional[DataNodeId] = None,
        name: Optional[str] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        last_edit_date: Optional[datetime] = None,
        edits: List[Edit] = None,
        version: str = None,
        validity_period: Optional[timedelta] = None,
        edit_in_progress: bool = False,
        editor_id: Optional[str] = None,
        editor_expiration_date: Optional[datetime] = None,
        properties: Dict = None,
    ):
        if properties is None:
            properties = {}
        required = self._REQUIRED_PROPERTIES
        if missing := set(required) - set(properties.keys()):
            raise MissingRequiredProperty(
                f"The following properties " f"{', '.join(x for x in missing)} were not informed and are required"
            )

        self._check_custom_document(properties[self._CUSTOM_DOCUMENT_PROPERTY])

        super().__init__(
            config_id,
            scope,
            id,
            name,
            owner_id,
            parent_ids,
            last_edit_date,
            edits,
            version or _VersionManagerFactory._build_manager()._get_latest_version(),
            validity_period,
            edit_in_progress,
            editor_id,
            editor_expiration_date,
            **properties,
        )

        mongo_client = _connect_mongodb(
            db_host=properties.get(self.__DB_HOST_KEY, self.__DB_HOST_DEFAULT),
            db_port=properties.get(self.__DB_PORT_KEY, self.__DB_PORT_DEFAULT),
            db_username=properties.get(self.__DB_USERNAME_KEY, ""),
            db_password=properties.get(self.__DB_PASSWORD_KEY, ""),
            db_driver=properties.get(self.__DB_DRIVER_KEY, ""),
            db_extra_args=frozenset(properties.get(self.__DB_EXTRA_ARGS_KEY, {}).items()),
        )
        self.collection = mongo_client[properties.get(self.__DB_NAME_KEY, "")][
            properties.get(self.__COLLECTION_KEY, "")
        ]

        self.custom_document = properties[self._CUSTOM_DOCUMENT_PROPERTY]

        self._decoder = self._default_decoder
        custom_decoder = getattr(self.custom_document, "decode", None)
        if callable(custom_decoder):
            self._decoder = custom_decoder

        self._encoder = self._default_encoder
        custom_encoder = getattr(self.custom_document, "encode", None)
        if callable(custom_encoder):
            self._encoder = custom_encoder

        if not self._last_edit_date:
            self._last_edit_date = datetime.now()

        self._TAIPY_PROPERTIES.update(
            {
                self.__COLLECTION_KEY,
                self.__DB_NAME_KEY,
                self._CUSTOM_DOCUMENT_PROPERTY,
                self.__DB_USERNAME_KEY,
                self.__DB_PASSWORD_KEY,
                self.__DB_HOST_KEY,
                self.__DB_PORT_KEY,
                self.__DB_DRIVER_KEY,
                self.__DB_EXTRA_ARGS_KEY,
            }
        )

    def _check_custom_document(self, custom_document):
        if not isclass(custom_document):
            raise InvalidCustomDocument(
                f"Invalid custom document of {custom_document}. Only custom class are supported."
            )

    @classmethod
    def storage_type(cls) -> str:
        return cls.__STORAGE_TYPE

    def _read(self):
        cursor = self._read_by_query()

        return list(map(lambda row: self._decoder(row), cursor))

    def _read_by_query(self):
        """Query from a Mongo collection, exclude the _id field"""

        return self.collection.find()

    def _write(self, data) -> None:
        """Check data against a collection of types to handle insertion on the database."""

        if not isinstance(data, list):
            data = [data]

        if len(data) == 0:
            self.collection.drop()
            return

        if isinstance(data[0], dict):
            self._insert_dicts(data)
        else:
            self._insert_dicts([self._encoder(row) for row in data])

    def _insert_dicts(self, data: List[Dict]) -> None:
        """
        :param data: a list of dictionaries

        This method will overwrite the data contained in a list of dictionaries into a collection.
        """
        self.collection.drop()
        self.collection.insert_many(data)

    def _default_decoder(self, document: Dict) -> Any:
        """Decode a Mongo dictionary to a custom document object for reading.

        Args:
            document (Dict): the document dictionary return by Mongo query.

        Returns:
            A custom document object.
        """
        return self.custom_document(**document)

    def _default_encoder(self, document_object: Any) -> Dict:
        """Encode a custom document object to a dictionary for writing to MongoDB.

        Args:
            document_object: the custom document class.

        Returns:
            The document dictionary.
        """
        return document_object.__dict__



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime
from typing import Dict, List, Optional, Set, Union

from .._repository._abstract_converter import _AbstractConverter
from .._version._utils import _migrate_entity
from ..common import _utils
from ..cycle._cycle_manager_factory import _CycleManagerFactory
from ..cycle.cycle import Cycle, CycleId
from ..data.data_node import DataNode, DataNodeId
from ..scenario._scenario_model import _ScenarioModel
from ..scenario.scenario import Scenario
from ..task.task import Task, TaskId


class _ScenarioConverter(_AbstractConverter):
    @classmethod
    def _entity_to_model(cls, scenario: Scenario) -> _ScenarioModel:
        sequences: Dict[str, Dict[str, Union[List[TaskId], Dict, List]]] = {}
        for p_name, sequence_data in scenario._sequences.items():
            sequences[p_name] = {
                Scenario._SEQUENCE_TASKS_KEY: [
                    t.id if isinstance(t, Task) else t for t in sequence_data.get("tasks", [])
                ],
                Scenario._SEQUENCE_PROPERTIES_KEY: sequence_data.get("properties", {}),
                Scenario._SEQUENCE_SUBSCRIBERS_KEY: _utils._fcts_to_dict(sequence_data.get("subscribers", [])),
            }

        return _ScenarioModel(
            id=scenario.id,
            config_id=scenario.config_id,
            tasks=[task.id if isinstance(task, Task) else TaskId(str(task)) for task in list(scenario._tasks)],
            additional_data_nodes=[
                dn.id if isinstance(dn, DataNode) else DataNodeId(str(dn))
                for dn in list(scenario._additional_data_nodes)
            ],
            properties=scenario._properties.data,
            creation_date=scenario._creation_date.isoformat(),
            primary_scenario=scenario._primary_scenario,
            subscribers=_utils._fcts_to_dict(scenario._subscribers),
            tags=list(scenario._tags),
            version=scenario._version,
            cycle=scenario._cycle.id if scenario._cycle else None,
            sequences=sequences if sequences else None,
        )

    @classmethod
    def _model_to_entity(cls, model: _ScenarioModel) -> Scenario:
        tasks: Union[Set[TaskId], Set[Task], Set] = set()
        if model.tasks:
            tasks = set(model.tasks)
        if model.sequences:
            for sequence_name, sequence_data in model.sequences.items():
                if subscribers := sequence_data.get(Scenario._SEQUENCE_SUBSCRIBERS_KEY):
                    model.sequences[sequence_name][Scenario._SEQUENCE_SUBSCRIBERS_KEY] = [
                        _utils._Subscriber(_utils._load_fct(it["fct_module"], it["fct_name"]), it["fct_params"])
                        for it in subscribers
                    ]

        scenario = Scenario(
            scenario_id=model.id,
            config_id=model.config_id,
            tasks=tasks,
            additional_data_nodes=set(model.additional_data_nodes),
            properties=model.properties,
            creation_date=datetime.fromisoformat(model.creation_date),
            is_primary=model.primary_scenario,
            tags=set(model.tags),
            cycle=cls.__to_cycle(model.cycle),
            subscribers=[
                _utils._Subscriber(_utils._load_fct(it["fct_module"], it["fct_name"]), it["fct_params"])
                for it in model.subscribers
            ],
            version=model.version,
            sequences=model.sequences,
        )
        return _migrate_entity(scenario)

    @staticmethod
    def __to_cycle(cycle_id: Optional[CycleId] = None) -> Optional[Cycle]:
        return _CycleManagerFactory._build_manager()._get(cycle_id) if cycle_id else None



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Dict

from .._repository._abstract_converter import _AbstractConverter
from ..common import _utils
from ..task.task import Task
from .sequence import Sequence


class _SequenceConverter(_AbstractConverter):
    _SEQUENCE_MODEL_ID_KEY = "id"
    _SEQUENCE_MODEL_OWNER_ID_KEY = "owner_id"
    _SEQUENCE_MODEL_PARENT_IDS_KEY = "parent_ids"
    _SEQUENCE_MODEL_PROPERTIES_KEY = "properties"
    _SEQUENCE_MODEL_TASKS_KEY = "tasks"
    _SEQUENCE_MODEL_SUBSCRIBERS_KEY = "subscribers"
    _SEQUENCE_MODEL_VERSION_KEY = "version"

    @classmethod
    def _entity_to_model(cls, sequence: Sequence) -> Dict:
        return {
            "id": sequence.id,
            "owner_id": sequence.owner_id,
            "parent_ids": list(sequence._parent_ids),
            "properties": sequence._properties.data,
            "tasks": cls.__to_task_ids(sequence._tasks),
            "subscribers": _utils._fcts_to_dict(sequence._subscribers),
            "version": sequence._version,
        }

    @staticmethod
    def __to_task_ids(tasks):
        return [t.id if isinstance(t, Task) else t for t in tasks]
